"""
Dear nT analyser,
if you want to complain please contact:
    chiara@physik.uzh.ch,
    gvolta@physik.uzh.ch,
    kazama@isee.nagoya-u.ac.jp
    torben.flehmke@fysik.su.se
"""

from immutabledict import immutabledict
import numba
import numpy as np
import scipy.stats as sps
import strax
import straxen

# This makes sure shorthands for only the necessary functions
# are made available under straxen.[...]
export, __all__ = strax.exporter()

channel_list = [i for i in range(494)]


@export
class LEDCalibration(strax.Plugin):
    """LEDCalibration Plugin for PMT LED signal analysis.

    This plugin processes raw PMT data and extracts various calibration
    parameters related to LED signals and noise. It is used during PMT
    calibration runs to measure LED pulse amplitudes, noise characteristics,
    and other relevant metrics. The analysis assumes a fixed LED window and
    computes parameters for both LED-on and LED-off runs.

    **Key Features:**
        - Extracts LED and noise amplitudes.
        - Computes the area of LED pulses using multiple integration windows
          for noise mitigation.
        - Identifies triggered intervals based on hits.

    **Returned Variables:**
        - `area`: Area under the LED pulse, averaged over multiple integration windows.
        - `area_noise`: Area in a noise window far from the LED signal.
        - `amplitude_led`: Peak amplitude of the LED signal.
        - `amplitude_noise`: Peak amplitude in a noise window.
        - `channel`: PMT channel identifier.
        - `time`: Start time of the interval (ns since UNIX epoch).
        - `dt`: Time resolution in nanoseconds.
        - `length`: Length of the interval in samples.
        - `triggered`: Boolean indicating if a hit was found in the record.
        - `hit_position`: Sample index of the hit defining the window position.
        - `integration_window`: Integration window used for area calculations.
        - `baseline`: Baseline of the record.

    """

    __version__ = "0.3.2"

    depends_on = "raw_records"
    data_kind = "led_cal"
    compressor = "zstd"
    parallel = "process"
    rechunk_on_save = False

    run_doc = straxen.URLConfig(
        default="run_doc://comments?run_id=plugin.run_id",
        infer_type=False,
        help=(
            "Comments dictionary from the run metadata. "
            "It is used for discriminate between LED on and LED "
            "off runs."
        ),
    )

    default_run_comments = straxen.URLConfig(
        default=["auto, SC user: ", "pulserScript: "],
        type=list,
        help=("List of default comments the automatic script for runs in PMT calibration. "),
    )

    noise_run_comments = straxen.URLConfig(
        default=["SPE_calibration_step0", "Gain_calibration_step3"],
        type=list,
        help=("List of comments for noise runs in PMT calibration. "),
    )

    led_cal_record_length = straxen.URLConfig(
        default=160, infer_type=False, help="Length (samples) of one record without 0 padding."
    )

    baseline_window = straxen.URLConfig(
        default=(0, 40),
        infer_type=False,
        help="Window (samples) for baseline calculation.",
    )

    minimum_led_position = straxen.URLConfig(
        default=60,
        infer_type=False,
        help=(
            "The minimum sample index to consider for LED hits. Hits before this sample are "
            "ignored."
        ),
    )

    fixed_position = straxen.URLConfig(
        default=88,
        infer_type=False,
        help=(
            "Fixed ADC sample upon which the integration window is defined. "
            "This is used as default when no hits or less than a certain amount "
            "are identified."
        ),
    )

    led_hit_extension = straxen.URLConfig(
        default=(-8, 32),
        infer_type=False,
        help="The extension around the LED hit to integrate.",
    )

    area_averaging_length = straxen.URLConfig(
        default=7,
        infer_type=False,
        help=(
            "The total length of the averaging window for the area calculation. "
            "To mitigate a possiple bias from noise, the area is integrated multiple times with "
            "sligntly different window lengths and then averaged. area_averaging_length should "
            "be divisible by area_averaging_step."
        ),
    )

    area_averaging_step = straxen.URLConfig(
        default=1,
        infer_type=False,
        help=(
            "The step size used for the different windows, averaged for the area calculation. "
            "To mitigate a possiple bias from noise, the area is integrated multiple times with "
            "sligntly different window lengths and then averaged. area_averaging_length should "
            "be divisible by area_averaging_step."
        ),
    )

    fixed_led_window = straxen.URLConfig(
        default=(78, 142),
        infer_type=False,
        help="Window (samples) where we expect the signal in LED calibration",
    )

    noise_window = straxen.URLConfig(
        default=(1, 65), infer_type=False, help="Window (samples) to analysis the noise"
    )

    channel_list = straxen.URLConfig(
        default=(tuple(channel_list)),
        infer_type=False,
        help="List of PMTs. Defalt value: all the PMTs",
    )

    led_cal_hit_min_height_over_noise = straxen.URLConfig(
        default=6,
        infer_type=False,
        help=("Minimum hit amplitude in numbers of baseline_rms above baseline. "),
    )

    dtype = strax.interval_dtype + [
        (("Area averaged in integration windows", "area"), np.float32),
        (("Area averaged in noise integration windows", "area_noise"), np.float32),
        (("Amplitude in LED window", "amplitude_led"), np.float32),
        (("Amplitude in off LED window", "amplitude_noise"), np.float32),
        (("Whether there was a hit found in the record", "triggered"), bool),
        (("Sample index of the hit that defines the window position", "hit_position"), np.uint8),
        (("Window used for integration", "integration_window"), np.uint8, (2,)),
        (("Baseline from the record", "baseline"), np.float32),
    ]

    def compute(self, raw_records):
        """The data for LED calibration are build for those PMT which belongs to channel list.

        This is used for the different ligh levels. As default value all the PMTs are considered.

        """

        self.is_led_on = is_the_led_on(
            self.run_doc, self.default_run_comments, self.noise_run_comments
        )

        mask = np.where(np.in1d(raw_records["channel"], self.channel_list))[0]
        raw_records_active_channels = raw_records[mask]
        records = get_records(
            raw_records_active_channels, self.baseline_window, self.led_cal_record_length
        )
        del raw_records_active_channels, raw_records

        temp = np.zeros(len(records), dtype=self.dtype)
        strax.copy_to_buffer(records, temp, "_recs_to_temp_led")

        led_windows, triggered = get_led_windows(
            records,
            self.minimum_led_position,
            self.fixed_position,
            self.is_led_on,
            self.led_hit_extension,
            self.led_cal_hit_min_height_over_noise,
            self.led_cal_record_length,
            self.area_averaging_length,
        )

        # The dynamic window algorithm gives
        # unexpected results. Therefore, we
        # decided to use an enlarged fixed window.
        _fixed_led_window = np.tile(self.fixed_led_window, (len(temp), 1))
        on, off = get_amplitude(records, _fixed_led_window, self.noise_window)
        temp["amplitude_led"] = on["amplitude"]
        temp["amplitude_noise"] = off["amplitude"]

        area = get_area(
            records, _fixed_led_window, self.area_averaging_length, self.area_averaging_step
        )
        temp["area"] = area["area"]

        _noise_window = np.tile(self.noise_window, (len(temp), 1))
        area = get_area(
            records, _noise_window, self.area_averaging_length, self.area_averaging_step
        )
        temp["area_noise"] = area["area"]

        temp["triggered"] = triggered
        # The hit position retrieves the result
        # of the hit-finding algorithm.
        temp["hit_position"] = led_windows[:, 0] - self.led_hit_extension[0]
        # The current version uses as integration window
        # an hardcoded fixed window
        temp["integration_window"] = _fixed_led_window
        temp["baseline"] = records["baseline"]
        return temp


def is_the_led_on(run_doc, default_run_comments, noise_run_comments):
    """Utilizing the run database metadata to determine whether the run ID corresponds to LED on or
    LED off runs.

    The LED off, or noise runs, are identified by having 'Gain_calibration_step3' or
    'SPE_calibration_step0' in the comment.

    """
    # Check if run_doc is a list with a dictionary
    if isinstance(run_doc, list) and isinstance(run_doc[0], dict):
        # Extract the dictionary
        doc = run_doc[0]

        # Check if the required keys are present
        required_keys = {"user", "date", "comment"}
        if all(key in doc for key in required_keys):
            # Check if 'comment' contains any of the noise run comments
            comment = doc["comment"]

            # Check if the comment matches the expected pattern
            if not all(x in comment for x in default_run_comments):
                raise ValueError("The comment does not match the expected pattern.")

            if any(noise_comment in comment for noise_comment in noise_run_comments):
                return False
            else:
                return True
        else:
            raise ValueError("The dictionary does not contain the required keys.")
    else:
        raise ValueError("The input is not a list with a single dictionary.")


def get_records(raw_records, baseline_window, led_cal_record_length):
    """Determine baseline as the average of the first baseline_samples of each pulse.

    Subtract the pulse float(data) from baseline.

    """

    record_length_padded = np.shape(raw_records.dtype["data"])[0]

    _dtype = strax.interval_dtype + [
        (
            (
                "Length of pulse to which the record belongs (without zero-padding)",
                "pulse_length",
            ),
            np.int32,
        ),
        (("Fragment number in the pulse", "record_i"), np.int16),
        (
            ("Baseline in ADC counts. data = int(baseline) - data_orig", "baseline"),
            np.float32,
        ),
        (
            ("Baseline RMS in ADC counts. data = baseline - data_orig", "baseline_rms"),
            np.float32,
        ),
        (
            ("Waveform data in raw ADC counts with 0 padding", "data"),
            np.float32,
            (record_length_padded,),
        ),
    ]

    records = np.zeros(len(raw_records), dtype=_dtype)
    strax.copy_to_buffer(raw_records, records, "_rr_to_r_led")

    mask = np.where((records["record_i"] == 0) & (records["length"] == led_cal_record_length))[0]
    records = records[mask]
    bl = records["data"][:, baseline_window[0] : baseline_window[1]].mean(axis=1)
    rms = records["data"][:, baseline_window[0] : baseline_window[1]].std(axis=1)
    records["data"][:, :led_cal_record_length] = (
        -1.0 * (records["data"][:, :led_cal_record_length].transpose() - bl[:]).transpose()
    )
    records["baseline"] = bl
    records["baseline_rms"] = rms
    return records


def get_led_windows(
    records,
    minimum_led_position,
    fixed_position,
    is_led_on,
    led_hit_extension,
    hit_min_height_over_noise,
    record_length,
    area_averaging_length,
):
    """Search for hits in the records, if a hit is found, return an interval around the hit given by
    led_hit_extension. If no hit is found in the record, return the default window.

    :param records: Array of the records to search for LED hits.
    :param minimum_led_position: The minimum simple index of the LED hits. Hits before this sample
        are ignored.
    :param is_led_on: Fetch from the run database. It is used for discriminate between LED on and
        LED off runs.
    :param fixed_position: Fixed ADC sample upon which the integration window is defined. Used if no
        hits are identified
    :param led_hit_extension: The integration window around the first hit found to use. A tuple of
        form (samples_before, samples_after) the first LED hit.
    :param hit_min_amplitude: Minimum amplitude of the signal to be considered a hit.
    :param hit_min_height_over_noise: Minimum height of the signal over noise to be considered a
        hit. :return (len(records), 2) array: Integration window for each record
    :param record_length: The length of one led_calibration record
    :param area_averaging_length: The length (samples) of the window to do the averaging on.

    """
    if len(records) == 0:  # If input is empty, return empty arrays of correct shape
        return np.empty((0, 2), dtype=np.int64), np.empty(0, dtype=bool)

    hits = strax.find_hits(
        records,
        min_amplitude=0,  # Always use the height over noise threshold.
        min_height_over_noise=hit_min_height_over_noise,
    )

    maximum_led_position = record_length - area_averaging_length - led_hit_extension[1]
    hits = hits[hits["left"] >= minimum_led_position]
    # Check if the records are sorted properly by 'record_i' first and 'time' second and sort them
    # if they are not
    record_i = hits["record_i"]
    time = hits["time"]
    if not (
        np.all(
            (record_i[:-1] < record_i[1:])
            | ((record_i[:-1] == record_i[1:]) & (time[:-1] <= time[1:]))
        )
    ):
        hits.sort(order=["record_i", "time"])

    # If there are not hits in the chunk or if
    # the run is a nosie run, with LED off,
    # the integration window is defined beased on a
    # hard-coded ADC sample
    if (not is_led_on) or (len(hits) == 0):
        default_hit_position = fixed_position
    else:
        default_hit_position = sps.mode(hits["left"])[0]

        if isinstance(default_hit_position, np.ndarray):
            default_hit_position = default_hit_position[0]

        if default_hit_position > maximum_led_position:
            default_hit_position = maximum_led_position

    triggered = np.zeros(len(records), dtype=bool)

    default_windows = np.tile(default_hit_position + np.array(led_hit_extension), (len(records), 1))
    # Once we take the mode of the hits arrival time
    # the hits are once more iterated and the
    # default_hit_position is overwitten we the
    # individual hit arrival time
    # what is then the point of doing this?
    # Giovanni 16/12/2024
    return _get_led_windows(
        hits, default_windows, led_hit_extension, maximum_led_position, triggered
    )


@numba.jit(nopython=True)
def _get_led_windows(hits, default_windows, led_hit_extension, maximum_led_position, triggered):
    windows = default_windows
    last = -1

    for hit in hits:
        if hit["record_i"] == last:
            continue  # If there are multiple hits in one record, ignore after the first

        triggered[hit["record_i"]] = True

        hit_left = hit["left"]
        # Limit the position of the window so it stays inside the record.
        if hit_left > maximum_led_position:
            hit_left = maximum_led_position

        left = hit_left + led_hit_extension[0]
        right = hit_left + led_hit_extension[1]

        windows[hit["record_i"]] = np.array([left, right])
        last = hit["record_i"]

    return windows, triggered


_on_off_dtype = np.dtype([("channel", "int16"), ("amplitude", "float32")])


@numba.jit(nopython=True)
def get_amplitude(records, led_windows, noise_window):
    """Needed for the SPE computation.

    Get the maximum of the signal in two different regions, one where there is no signal, and one
    where there is.

    :param records: Array of records
    :param ndarray led_windows : 2d array of shape (len(records), 2) with the window to use as the
        signal on area for each record. Inclusive left boundary and exclusive right boundary.
    :param tuple noise_window: Tuple with the window, used for the signal off area for all records.
    :return ndarray ons: 1d array of length len(records). The maximum amplitude in the led window
        area for each record.
    :return ndarray offs: 1d array of length len(records). The maximum amplitude in the noise area
        for each record.

    """
    ons = np.zeros(len(records), dtype=_on_off_dtype)
    offs = np.zeros(len(records), dtype=_on_off_dtype)

    for i, record in enumerate(records):
        ons[i]["channel"] = record["channel"]
        offs[i]["channel"] = record["channel"]

        ons[i]["amplitude"] = np.max(record["data"][led_windows[i, 0] : led_windows[i, 1]])
        offs[i]["amplitude"] = np.max(record["data"][noise_window[0] : noise_window[1]])

    return ons, offs


_area_dtype = np.dtype([("channel", "int16"), ("area", "float32")])


@numba.jit(nopython=True)
def get_area(records, led_windows, area_averaging_length, area_averaging_step):
    """Needed for the gain computation.

    Integrate the record in the defined window area. To reduce the effects of the noise, this is
    done with 6 different window lengths, which are then averaged.

    :param records: Array of records
    :param ndarray led_windows : 2d array of shape (len(records), 2) with the window to use as the
        integration boundaries.
    :param area_averaging_length: The total length in records of the window over which to do the
        averaging of the areas.
    :param area_averaging_step: The increase in length for each step of the averaging.
    :return ndarray area: 1d array of length len(records) with the averaged integrated areas for
        each record.

    """
    area = np.zeros(len(records), dtype=_area_dtype)
    end_pos = np.arange(0, area_averaging_length + area_averaging_step, area_averaging_step)

    for i, record in enumerate(records):
        area[i]["channel"] = record["channel"]
        for right in end_pos:
            area[i]["area"] += np.sum(record["data"][led_windows[i, 0] : led_windows[i, 1] + right])

        area[i]["area"] /= float(len(end_pos))

    return area


@export
class nVetoExtTimings(strax.Plugin):
    """Plugin which computes the time difference `delta_time` from pulse timing of `hitlets_nv` to
    start time of `raw_records` which belong the `hitlets_nv`.

    They are used as the external trigger timings.

    """

    __version__ = "0.0.1"

    depends_on = ("raw_records_nv", "hitlets_nv")
    provides = "ext_timings_nv"
    data_kind = "hitlets_nv"

    compressor = "zstd"

    channel_map = straxen.URLConfig(
        track=False,
        type=immutabledict,
        help="immutabledict mapping subdetector to (min, max) channel number.",
    )

    def infer_dtype(self):
        dtype = []
        dtype += strax.time_dt_fields
        dtype += [
            (("Delta time from trigger timing [ns]", "delta_time"), np.int16),
            (
                ("Index to which pulse (not record) the hitlet belongs to.", "pulse_i"),
                np.int32,
            ),
        ]
        return dtype

    def setup(self):
        self.nv_pmt_start = self.channel_map["nveto"][0]
        self.nv_pmt_stop = self.channel_map["nveto"][1] + 1

    def compute(self, hitlets_nv, raw_records_nv):
        rr_nv = raw_records_nv[raw_records_nv["record_i"] == 0]
        pulses = np.zeros(len(rr_nv), dtype=self.pulse_dtype())
        pulses["time"] = rr_nv["time"]
        pulses["endtime"] = rr_nv["time"] + rr_nv["pulse_length"] * rr_nv["dt"]
        pulses["channel"] = rr_nv["channel"]

        ext_timings_nv = np.zeros_like(hitlets_nv, dtype=self.dtype)
        ext_timings_nv["time"] = hitlets_nv["time"]
        ext_timings_nv["length"] = hitlets_nv["length"]
        ext_timings_nv["dt"] = hitlets_nv["dt"]
        self.calc_delta_time(
            ext_timings_nv, pulses, hitlets_nv, self.nv_pmt_start, self.nv_pmt_stop
        )

        return ext_timings_nv

    @staticmethod
    def pulse_dtype():
        pulse_dtype = []
        pulse_dtype += strax.time_fields
        pulse_dtype += [(("PMT channel", "channel"), np.int16)]
        return pulse_dtype

    @staticmethod
    def calc_delta_time(ext_timings_nv_delta_time, pulses, hitlets_nv, nv_pmt_start, nv_pmt_stop):
        """Numpy access with fancy index returns copy, not view This for-loop is required to
        substitute in one by one."""
        hitlet_index = np.arange(len(hitlets_nv))
        pulse_index = np.arange(len(pulses))
        for ch in range(nv_pmt_start, nv_pmt_stop):
            mask_hitlets_in_channel = hitlets_nv["channel"] == ch
            hitlet_in_channel_index = hitlet_index[mask_hitlets_in_channel]

            mask_pulse_in_channel = pulses["channel"] == ch
            pulse_in_channel_index = pulse_index[mask_pulse_in_channel]

            hitlets_in_channel = hitlets_nv[hitlet_in_channel_index]
            pulses_in_channel = pulses[pulse_in_channel_index]
            hit_in_pulse_index = strax.fully_contained_in(hitlets_in_channel, pulses_in_channel)
            for h_i, p_i in zip(hitlet_in_channel_index, hit_in_pulse_index):
                if p_i == -1:
                    continue
                res = ext_timings_nv_delta_time[h_i]

                res["delta_time"] = (
                    hitlets_nv[h_i]["time"]
                    + hitlets_nv[h_i]["time_amplitude"]
                    - pulses_in_channel[p_i]["time"]
                )
                res["pulse_i"] = pulse_in_channel_index[p_i]



import numpy as np
import numba
import strax
import straxen
from ._peaklet_positions_base import PeakletPositionsBase

export, __all__ = strax.exporter()


@export
class PeakletPositionsCNF(PeakletPositionsBase):
    """Conditional Normalizing Flow for position reconstruction.

    This plugin reconstructs the position of peaklets using a conditional normalizing flow model.
    It provides x and y coordinates of the reconstructed position, along with uncertainty contours
    and uncertainty estimates in r and theta. For information on the model, see note_.

    .. _note: https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:juehang:flow_posrec_proposal_sr2  # noqa: E501

    Configuration options:
    - min_reconstruction_area: Minimum area (PE) required for reconstruction
    - n_poly: Size of the uncertainty contour
    - sig: Confidence level of the contour
    - log_area_scale: Scaling parameter for log area
    - n_top_pmts: Number of top PMTs
    - cnf_pred_function: Path to the compiled JAX function for predictions

    """

    __version__ = "0.0.1"
    child_plugin = True
    algorithm = "cnf"
    provides = "peaklet_positions_cnf"

    n_poly = straxen.URLConfig(
        default=16,
        infer_type=False,
        help="Size of uncertainty contour",
    )

    N_chunk_max = straxen.URLConfig(
        default=4096,
        infer_type=False,
        help="Maximum size of chunk for vectorised JAX function",
    )

    sig = straxen.URLConfig(
        default=0.393,
        infer_type=False,
        help="Confidence level of contour",
    )

    log_area_scale = straxen.URLConfig(
        default=10,
        infer_type=False,
        help="Scaling parameter for log area",
    )

    cnf_pred_function = straxen.URLConfig(
        default=(
            "jax://resource://xedocs://posrec_models"
            "?kind=cnf&attr=value&n_poly=plugin.n_poly&sig=plugin.sig&fmt=abs_path"
            "&version=ONLINE&run_id=plugin.run_id"
        ),
        help="Compiled JAX function",
    )

    @staticmethod
    def calculate_theta_diff(theta_array, avg_theta):
        """Calculate the difference between maximum and minimum angles from an array of angles by
        normalizing the angular difference into the range [0, 2π).

        Parameters:
            theta_array : np.ndarray
            A 2D numpy array where each row represents a set of angles in radians.
            avg_theta : np.ndarray
            A 1D numpy array representing the average angle in radians for each
            row in `theta_array`.

        Returns:
            theta_diff : np.ndarray
            A 1D numpy array with the difference between the maximum and minimum angles in radians
            for each row in `theta_array`

        """
        # Correction to handle circular nature of angles
        theta_array_shift = (theta_array - avg_theta[..., np.newaxis] + np.pi) % (2 * np.pi)
        theta_min = np.min(theta_array_shift, axis=1)
        theta_max = np.max(theta_array_shift, axis=1)
        theta_diff = theta_max - theta_min

        return theta_diff

    def infer_dtype(self):
        """Define the data type for the output.

        Returns:
            dtype: Numpy dtype for the output array

        """
        dtype = [
            (
                (
                    f"Reconstructed {self.algorithm} S2 X position (cm), uncorrected",
                    f"x_{self.algorithm}",
                ),
                np.float32,
            ),
            (
                (
                    f"Reconstructed {self.algorithm} S2 Y position (cm), uncorrected",
                    f"y_{self.algorithm}",
                ),
                np.float32,
            ),
            (
                ("Position uncertainty contour (cm)", f"position_contour_{self.algorithm}"),
                np.float32,
                (self.n_poly + 1, 2),
            ),
            (
                (
                    "Area in position uncertainty contour (cm^2)",
                    f"position_contour_area_{self.algorithm}",
                ),
                np.float32,
            ),
            (("Position uncertainty in r (cm)", f"r_uncertainty_{self.algorithm}"), np.float32),
            (
                ("Position uncertainty in theta (rad)", f"theta_uncertainty_{self.algorithm}"),
                np.float32,
            ),
        ]
        dtype += strax.time_fields
        return dtype

    def vectorized_prediction_chunk(self, flow_condition):
        """Compute predictions for a chunk of data.

        Args:
            flow_condition: Input data for the flow model

        Returns:
            xy: Predicted x and y coordinates
            contour: Uncertainty contours

        """
        N_entries = flow_condition.shape[0]
        if N_entries > self.N_chunk_max:
            raise ValueError("Chunk greater than max size")
        else:
            inputs = np.zeros((self.N_chunk_max, self.n_top_pmts + 1))
            inputs[:N_entries] = flow_condition
            xy, contour = self.cnf_pred_function(inputs)
            return xy[:N_entries], contour[:N_entries]

    def prediction_loop(self, flow_condition):
        """Compute predictions for arbitrary-size inputs using a loop.

        Args:
            flow_condition: Input data for the flow model

        Returns:
            xy: Predicted x and y coordinates
            contour: Uncertainty contours

        """
        N_entries = flow_condition.shape[0]
        if N_entries <= self.N_chunk_max:
            return self.vectorized_prediction_chunk(flow_condition)
        N_chunks = N_entries // self.N_chunk_max

        xy_list = []
        contour_list = []
        for i in range(N_chunks):
            xy, contour = self.vectorized_prediction_chunk(
                flow_condition[i * self.N_chunk_max : (i + 1) * self.N_chunk_max]
            )
            xy_list.append(xy)
            contour_list.append(contour)

        if N_chunks * self.N_chunk_max < N_entries:
            xy, contour = self.vectorized_prediction_chunk(
                flow_condition[(i + 1) * self.N_chunk_max :]
            )
            xy_list.append(xy)
            contour_list.append(contour)
        return np.concatenate(xy_list, axis=0), np.concatenate(contour_list, axis=0)

    @staticmethod
    @numba.njit(cache=True, nogil=True)
    def polygon_area(polygon):
        """Calculate and return the area of a polygon.

        The input is a 3D numpy array where the first dimension represents individual polygons, the
        second dimension represents vertices of the polygon, and the third dimension represents x
        and y coordinates of each vertex.

        """
        x = polygon[..., 0]
        y = polygon[..., 1]
        result = np.zeros(polygon.shape[0], dtype=np.float32)
        for i in range(x.shape[-1]):
            result += (x[..., i] * y[..., i - 1]) - (y[..., i] * x[..., i - 1])
        return 0.5 * np.abs(result)

    def compute(self, peaklets):
        """Compute the position reconstruction for the given peaklets.

        Args:
            peaklets: Input peaklet data

        Returns:
            result: Array with reconstructed positions and uncertainties

        """
        # Initialize result array
        result = np.ones(len(peaklets), dtype=self.dtype)
        # Set default values to NaN
        strax.set_nan_defaults(result)
        result["time"], result["endtime"] = peaklets["time"], strax.endtime(peaklets)

        # Keep large peaklets only
        peaklet_mask = peaklets["area"] > self.min_reconstruction_area
        if not np.sum(peaklet_mask):
            # Nothing to do, and .predict crashes on empty arrays
            return result

        # Prepare input data for the flow model
        area_per_channel_top = peaklets["area_per_channel"][peaklet_mask, : self.n_top_pmts]
        total_top_areas = np.sum(area_per_channel_top, axis=1)
        with np.errstate(divide="ignore", invalid="ignore"):
            flow_data = np.concatenate(
                [
                    area_per_channel_top / total_top_areas[..., np.newaxis],
                    np.log(total_top_areas[..., np.newaxis]) / self.log_area_scale,
                ],
                axis=1,
            )

        # Get position reconstruction
        xy, contours = self.prediction_loop(flow_data)

        # Write output to the result array
        result[f"x_{self.algorithm}"][peaklet_mask] = xy[:, 0]
        result[f"y_{self.algorithm}"][peaklet_mask] = xy[:, 1]
        result[f"position_contour_{self.algorithm}"][peaklet_mask] = contours
        result[f"position_contour_area_{self.algorithm}"][peaklet_mask] = self.polygon_area(
            result[f"position_contour_{self.algorithm}"][peaklet_mask]
        )

        # Calculate uncertainties in r and theta
        r_array = np.linalg.norm(contours, axis=2)
        r_min = np.min(r_array, axis=1)
        r_max = np.max(r_array, axis=1)

        theta_array = np.arctan2(contours[..., 1], contours[..., 0])

        avg_theta = np.arctan2(xy[:, 1], xy[:, 0])

        theta_diff = self.calculate_theta_diff(theta_array, avg_theta)

        result[f"r_uncertainty_{self.algorithm}"][peaklet_mask] = (r_max - r_min) / 2
        result[f"theta_uncertainty_{self.algorithm}"][peaklet_mask] = np.abs(theta_diff) / 2

        return result



import strax
from straxen.plugins.peaklets.peaklet_classification_som import som_additional_fields
from straxen.plugins.peaks.peak_basics_vanilla import PeakBasicsVanilla

export, __all__ = strax.exporter()


@export
class PeakBasicsSOM(PeakBasicsVanilla):
    """Adds SOM fields to peak basics to be propgated to event basics."""

    __version__ = "0.0.1"
    child_plugin = True

    def infer_dtype(self):
        dtype = super().infer_dtype()
        return dtype + som_additional_fields

    def compute(self, peaks):
        peak_basics = super().compute(peaks)
        fields_to_copy = strax.to_numpy_dtype(som_additional_fields).names
        strax.copy_to_buffer(peaks, peak_basics, "_copy_som_information", fields_to_copy)
        return peak_basics



import strax
import numpy as np
from immutabledict import immutabledict
import straxen

export, __all__ = strax.exporter()


@export
class OnlineMonitorNV(strax.Plugin):
    """Plugin to write data of nVeto detector to the online-monitor. Data that is written by this
    plugin should be small (~MB/chunk) to not overload the runs-database.

    This plugin takes 'hitlets_nv' and 'events_nv'. Although they are not strictly related, they are
    aggregated into a single data_type in order to minimize the number of documents in the online
    monitor.

    Produces 'online_monitor_nv' with info on the hitlets_nv and events_nv

    """

    __version__ = "0.0.4"

    depends_on = ("hitlets_nv", "events_nv")
    provides = "online_monitor_nv"
    data_kind = "online_monitor_nv"

    # Needed in case we make again an muVETO child.
    ends_with = "_nv"

    channel_map = straxen.URLConfig(
        track=False,
        type=immutabledict,
        help="immutabledict mapping subdetector to (min, max) channel number.",
    )

    events_area_bounds = straxen.URLConfig(
        type=tuple,
        default=(-0.5, 130.5),
        help="Boundaries area histogram of events_nv_area_per_chunk [PE]",
    )

    events_area_nbins = straxen.URLConfig(
        type=int,
        default=131,
        help="Number of bins of histogram of events_nv_area_per_chunk, defined value 1 PE/bin",
    )

    def infer_dtype(self):
        self.channel_range = self.channel_map["nveto"]
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_monitor_dtype(self.ends_with, self.n_channel, self.events_area_nbins)

    def compute(self, hitlets_nv, events_nv, start, end):
        # General setup
        res = np.zeros(1, dtype=self.dtype)
        res["time"] = start
        res["endtime"] = end

        # Count number of hitlets_nv per PMT
        hitlets_channel_count, _ = np.histogram(
            hitlets_nv["channel"],
            bins=self.n_channel,
            range=[self.channel_range[0], self.channel_range[1] + 1],
        )
        res[f"hitlets{self.ends_with}_per_channel"] = hitlets_channel_count

        # Count number of events_nv with coincidence cut
        res[f"events{self.ends_with}_per_chunk"] = len(events_nv)
        sel = events_nv["n_contributing_pmt"] >= 4
        res[f"events{self.ends_with}_4coinc_per_chunk"] = np.sum(sel)
        sel = events_nv["n_contributing_pmt"] >= 5
        res[f"events{self.ends_with}_5coinc_per_chunk"] = np.sum(sel)
        sel = events_nv["n_contributing_pmt"] >= 8
        res[f"events{self.ends_with}_8coinc_per_chunk"] = np.sum(sel)
        sel = events_nv["n_contributing_pmt"] >= 10
        res[f"events{self.ends_with}_10coinc_per_chunk"] = np.sum(sel)

        # Get histogram of events_nv_area per chunk
        events_area, bins_ = np.histogram(
            events_nv["area"], bins=self.events_area_nbins, range=self.events_area_bounds
        )
        res[f"events{self.ends_with}_area_per_chunk"] = events_area
        return res


def veto_monitor_dtype(veto_name: str = "_nv", n_pmts: int = 120, n_bins: int = 131) -> list:
    dtype = []
    dtype += strax.time_fields  # because mutable
    dtype += [
        (
            (f"hitlets{veto_name} per channel", f"hitlets{veto_name}_per_channel"),
            (np.int64, n_pmts),
        ),
        (
            (f"events{veto_name}_area per chunk", f"events{veto_name}_area_per_chunk"),
            np.int64,
            n_bins,
        ),
        ((f"events{veto_name} per chunk", f"events{veto_name}_per_chunk"), np.int64),
        (
            (f"events{veto_name} 4-coincidence per chunk", f"events{veto_name}_4coinc_per_chunk"),
            np.int64,
        ),
        (
            (f"events{veto_name} 5-coincidence per chunk", f"events{veto_name}_5coinc_per_chunk"),
            np.int64,
        ),
        (
            (f"events{veto_name} 8-coincidence per chunk", f"events{veto_name}_8coinc_per_chunk"),
            np.int64,
        ),
        (
            (f"events{veto_name} 10-coincidence per chunk", f"events{veto_name}_10coinc_per_chunk"),
            np.int64,
        ),
    ]
    return dtype



import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class EnergyEstimates(strax.Plugin):
    """Plugin which converts cS1 and cS2 into energies (from PE to KeVee)."""

    __version__ = "0.1.1"
    depends_on = "corrected_areas"
    dtype = [
        ("e_light", np.float32, "Energy in light signal [keVee]"),
        ("e_charge", np.float32, "Energy in charge signal [keVee]"),
        ("e_ces", np.float32, "Energy estimate [keVee]"),
    ] + strax.time_fields
    save_when = strax.SaveWhen.TARGET

    # config options don't double cache things from the resource cache!
    g1 = straxen.URLConfig(
        default="bodega://g1?bodega_version=v2",
        help="S1 gain in PE / photons produced",
    )
    g2 = straxen.URLConfig(
        default="bodega://g2?bodega_version=v2",
        help="S2 gain in PE / electrons produced",
    )
    lxe_w = straxen.URLConfig(default=13.7e-3, help="LXe work function in quanta/keV")

    def compute(self, events):
        el = self.cs1_to_e(events["cs1"])
        ec = self.cs2_to_e(events["cs2"])
        return dict(
            e_light=el,
            e_charge=ec,
            e_ces=el + ec,
            time=events["time"],
            endtime=strax.endtime(events),
        )

    def cs1_to_e(self, x):
        return self.lxe_w * x / self.g1

    def cs2_to_e(self, x):
        return self.lxe_w * x / self.g2



"""Dynamic holoviews-based waveform display Note imports are inside function, to keep 'import
straxen' free of holoviews."""

import numpy as np
import pandas as pd
import strax
import straxen

straxen._BOKEH_X_RANGE = None


@straxen.mini_analysis(requires=["hitlets_nv", "events_nv", "event_positions_nv"])
def plot_nveto_event_display(context, run_id, hitlets_nv, events_nv):
    nvd = straxen.nVETOEventDisplay(events_nv, hitlets_nv, run_id)
    return nvd.plot_event_display()


def seconds_from(t, t_reference, unit_conversion=int(1e9)):
    """Helper function which concerts times into relative times in specified unit.

    :param t: Time
    :param t_reference: Reference time e.g. start of an event or first peak in event.
    :param unit_conversion: Conversion factor for time units e.g. 10**3 for micro seconds.

    """
    return (t - t_reference) / unit_conversion


# Custom wheel zoom tool that only zooms in one dimension
def x_zoom_wheel():
    import bokeh.models

    return bokeh.models.WheelZoomTool(dimensions="width")


@straxen.mini_analysis(requires=["records"], hv_bokeh=True)
def hvdisp_plot_pmt_pattern(*, config, records, to_pe, array="bottom"):
    """Plot a PMT array, with colors showing the intensity of light observed in the time range.

    :param array: 'top' or 'bottom', array to show.

    """
    import holoviews as hv

    pmts = straxen.pmt_positions()
    areas = np.bincount(
        records["channel"], weights=records["area"] * to_pe[records["channel"]], minlength=len(pmts)
    )

    # Which PMTs should we include?
    m = pmts["array"] == array
    pmts = pmts[m].copy()
    pmts["area"] = areas[m]

    f = 1.08
    pmts = hv.Dataset(
        pmts,
        kdims=[
            hv.Dimension("x", unit="cm", range=(-straxen.tpc_r * f, straxen.tpc_r * f)),
            hv.Dimension("y", unit="cm", range=(-straxen.tpc_r * f, straxen.tpc_r * f)),
            hv.Dimension("i", range=(0, config["n_tpc_pmts"]), label="PMT number"),
            hv.Dimension("area", label="Area", unit="PE"),
        ],
    )
    pmts = pmts.to(hv.Points, vdims=["area", "i"], group="PMTPattern", label=array.capitalize())

    hv.opts.apply_groups(
        pmts,
        plot=dict(color_index=2, tools=["hover"], show_grid=False),
        style=dict(size=17, cmap="plasma"),
    )

    return pmts


def _records_to_points(*, records, to_pe, t_reference, config, unit_conversion=int(1e9)):
    """Return (holoviews.Points, time_stream) corresponding to records."""
    import holoviews as hv

    areas_r = records["area"] * to_pe[records["channel"]]

    # Create dataframe with record metadata
    df = pd.DataFrame(
        dict(
            area=areas_r,
            time=seconds_from(
                records["time"] + records["dt"] * records["length"] // 2,
                t_reference,
                unit_conversion=unit_conversion,
            ),
            length=records["length"],
            dt=records["dt"],
            channel=records["channel"],
        )
    )

    rec_points = hv.Points(
        df,
        kdims=[
            hv.Dimension("time", label="Time [µs]"),
            hv.Dimension("channel", label="PMT Channel", range=(0, config["n_tpc_pmts"])),
        ],
        vdims=[hv.Dimension("area", label="Area [pe]")],
    )

    time_stream = hv.streams.RangeX(source=rec_points)
    return rec_points, time_stream


@straxen.mini_analysis(requires=["records"], hv_bokeh=True)
def hvdisp_plot_records_2d(
    records,
    to_pe,
    config,
    t_reference,
    time_stream=None,
    tools=(x_zoom_wheel(), "xpan"),
    default_tools=("save", "pan", "box_zoom", "save", "reset"),
    plot_library="bokeh",
    hooks=(),
):
    """Plot records in a dynamic 2D histogram of (time, pmt)

    :param width: Plot width in pixels
    :param time_stream: holoviews rangex stream to use. If provided, we assume records is already
        converted to points (which hopefully is what the stream is derived from)
    :param tools: Tools to be used in the interactive plot. Only works with bokeh as plot library.
    :param plot_library: Default bokeh, library to be used for the plotting.
    :param width: With of the record matrix in pixel.
    :param hooks: Hooks to adjust plot settings.
    :return: datashader object, records holoview points, RangeX time stream of records.

    """
    shader, records, time_stream = _hvdisp_plot_records_2d(
        records,
        to_pe,
        config,
        t_reference,
        time_stream=time_stream,
        default_tools=default_tools,
        tools=tools,
        hooks=hooks,
        plot_library=plot_library,
    )
    shader = shader.opts(title="Time vs. Channel")
    return shader


def _hvdisp_plot_records_2d(
    records,
    to_pe,
    config,
    t_reference,
    event_range=(None, None),
    time_stream=None,
    default_tools=(),
    tools=(),
    hooks=(),
    plot_library="bokeh",
    unit_conversion=10**3,
):
    import holoviews as hv
    import holoviews.operation.datashader

    hv.extension(plot_library)

    if time_stream is None:
        # Records are still a dataframe, convert it to points
        records, time_stream = _records_to_points(
            records=records,
            to_pe=to_pe,
            t_reference=t_reference,
            unit_conversion=unit_conversion,
            config=config,
        )

    # Whether to show the toolbar or not:
    if tools:
        toolbar = "right"
    else:
        toolbar = None

    # Creating the plot:
    shader = hv.operation.datashader.dynspread(
        hv.operation.datashader.datashade(
            records,
            dynamic=True,
            x_range=event_range,
            y_range=(0, config["n_tpc_pmts"]),
            streams=[time_stream],
        ),
        threshold=0.1,
    )

    hv.opts.apply_groups(
        shader,
        plot=dict(
            aspect=4,
            responsive="width",
            hooks=list(hooks),
            toolbar=toolbar,
            tools=list(tools),
            default_tools=list(default_tools),
            fontsize={"labels": 12},
            show_grid=True,
        ),
    )

    return shader, records, time_stream


def plot_record_polygons(
    record_points,
    center_time=True,
    scaling=10**-3,
):
    """Plots record hv.Points as polygons for record matrix.

    :param record_points: Holoviews Points generated with _records_to_points.
    :param center_time: If true treats specified times as center times.
    :param scaling: Scale times scale by this factor. E.g. if times should be in µs use 10**-3.
    :return: hv.Polygons

    """
    import holoviews as hv

    df = record_points.data.copy()
    if not center_time:
        df.loc[:, "time"] = df.loc[:, "time"] + (df.loc[:, "length"] / 2 * df.loc[:, "dt"])

    data = _make_polys(df, scaling=scaling)
    polys = hv.Polygons(data, kdims=["time", "channel"], vdims=list(data[0].keys())[1:]).opts(
        color="area",
        aspect=4,
        responsive="width",
        line_width=0,
        cmap="viridis",
    )

    return polys


def _make_polys(df, scaling=1):
    """Function which converts the hitlet/record points into polygons for the hitlet/record
    matrix."""
    res = []
    for row_i in range(len(df)):
        data_dict = {}
        # Kdims are given by time and channel:
        t, ch = df.loc[row_i, ["time", "channel"]]
        w = df.loc[row_i, "length"] * df.loc[row_i, "dt"]
        data_dict[("time", "channel")] = _rectangle(t, ch, width=w * scaling)

        # Add other parameters to polygons for vdmis:
        keys = [key for key in df.columns.values if key not in ["time", "channel", "dt"]]
        for key in keys:
            data_dict[key] = df.loc[row_i, key]
        res.append(data_dict)
    return res


def _rectangle(time=0, channel=0, width=1.1, height=1):
    """Generates polygon box coordinates for record matrix.

    :param time: Center position of the record in time.
    :param channel: Center position of PMT channel. E.g channel 0 => 0.5
    :param width: Length of the record in µs.
    param height: Height of the record in "channel"-units.

    X,Y have to be the center of the polygon.
    Width and height are the full width and height in data coordinates.

    """
    width = width / 2
    height = height / 2
    return np.array(
        [
            (time - width, channel - height),
            (time + width, channel - height),
            (time + width, channel + height),
            (time - width, channel + height),
        ]
    )


def get_records_matrix_in_window(polys, x_range, time_slice=10):
    """Helper function which returns polygons for rendering when x_range is below the specified
    value.

    This function has to be applied to polygons e.g.:

    poly.apply(get_records_matrix_in_window, streams=[time_stream])

    :param polys: Holoviews Polygons
    :param x_range: x_range of the RangeX object.
    :param time_slice: Size of the time slice in [µs] when records should be drawn.

    """
    if x_range is None:
        # Needed since x_range is by default not defined when plotting
        # the first time.
        return polys.iloc[:0]
    if (x_range[1] - x_range[0]) < time_slice:
        # If x_range smaller than specified minimum return polys ->
        # render polys.
        inds = _in_window(polys.data, x_range)
        return polys.iloc[inds]
    return polys.iloc[:0]


def _in_window(polys, x_range):
    """Function which checks if a polygon is partially in x_range.

    :param polys: List of ordered dictionaries containing Polygon data.
    :param x_range: Range which should be tested.

    """
    res = []
    for ind, p in enumerate(polys):
        # Loop over polys, if partially in window keep index.
        if np.any((x_range[0] <= p["time"]) & (p["time"] < x_range[1])):
            res.append(ind)
    return res


@straxen.mini_analysis(requires=["peaks", "peak_basics"], hv_bokeh=True)
def hvdisp_plot_peak_waveforms(
    t_reference, time_range, peaks, width=600, show_largest=None, time_dim=None
):
    """Plot the sum waveforms of peaks. Holoviews time dimension; will create new one if not
    provided.

    :param width: Plot width in pixels
    :param show_largest: Maximum number of peaks to show
    :param time_dim:

    """
    import holoviews as hv

    if show_largest is not None and len(peaks) > show_largest:
        show_i = strax.stable_argsort(peaks["area"])[-show_largest::]
        peaks = peaks[show_i]

    curves = []
    for p in peaks:
        # label = {1: 's1', 2: 's2'}.get(
        #     p['type'], 'unknown')
        color = {1: "blue", 2: "green"}.get(p["type"], "k")

        # It's better to plot amplitude /time than per bin, since
        # sampling times are now variable
        y = p["data"][: p["length"]] / p["dt"]
        t_edges = np.arange(p["length"] + 1, dtype=np.int64)
        t_edges = t_edges * p["dt"] + p["time"]
        t_edges = seconds_from(t_edges, t_reference)

        # Make a 'step' plot. Unlike matplotlib's steps-mid,
        # this also analyses the final edges correctly
        t_ = np.zeros(2 * len(y))
        y_ = np.zeros(2 * len(y))
        t_[0::2] = t_edges[:-1]
        t_[1::2] = t_edges[1:]
        y_[0::2] = y
        y_[1::2] = y

        if time_dim is None:
            time_dim = hv.Dimension("time", label="Time", unit="sec")

        curves.append(
            hv.Curve(
                dict(time=t_, amplitude=y_),
                kdims=time_dim,
                vdims=hv.Dimension("amplitude", label="Amplitude", unit="PE/ns"),
                group="PeakSumWaveform",
            )
        )

        hv.opts.apply_groups(
            curves[-1],
            style=dict(color=color),
        )

    overlay = hv.Overlay(items=curves)
    hv.opts.apply_groups(
        overlay,
        plot=dict(width=width),
    )

    return overlay


def _range_plot(f, full_time_range, t_reference, **kwargs):
    # The **bla is needed to disable some holoviews check
    # on the arguments...
    def wrapped(x_range, **kwargzz):
        if len(kwargzz):
            raise RuntimeError(f"Passed superfluous kwargs {kwargzz}")
        if x_range is None:
            x_range = seconds_from(np.asarray(full_time_range), t_reference)

        # Deal with strange time ranges -- not sure how these arise?
        x_range = np.nan_to_num(x_range)
        if x_range[1] == x_range[0]:
            x_range[1] += 1

        return f(
            time_range=(t_reference + int(x_range[0] * 1e9), t_reference + int(x_range[1] * 1e9)),
            t_reference=t_reference,
            **kwargs,
        )

    return wrapped


@straxen.mini_analysis(requires=["records", "peaks", "peak_basics"], hv_bokeh=True)
def waveform_display(
    context,
    run_id,
    to_pe,
    time_range,
    t_reference,
    records,
    peaks,
    config,
    width=600,
    show_largest=None,
):
    """Plot a waveform overview display".

    :param width: Plot width in pixels.

    """
    import holoviews as hv

    records_points, time_stream = _records_to_points(
        records=records, to_pe=to_pe, t_reference=t_reference, config=config
    )

    time_v_channel = context.hvdisp_plot_records_2d(
        run_id=run_id,
        to_pe=to_pe,
        records=records_points,
        time_stream=time_stream,
        time_range=time_range,
        t_reference=t_reference,
        # We don't need to cut these further, records we get are already cut
        # to the plot's maximum range by the mini_analysis logic
        # and datashader does the online cutting / rebinning / zooming.
        # This is fortunate, since we omitted 'endtime' from records_points!
        time_selection="skip",
    )

    array_plot = {
        array: hv.DynamicMap(
            _range_plot(
                context.hvdisp_plot_pmt_pattern,
                run_id=run_id,
                to_pe=to_pe,
                records=records,
                full_time_range=time_range,
                t_reference=t_reference,
                time_selection="touching",
                array=array,
            ),
            streams=[time_stream],
        )
        for array in ("top", "bottom")
    }

    peak_wvs = hv.DynamicMap(
        _range_plot(
            context.hvdisp_plot_peak_waveforms,
            run_id=run_id,
            width=width,
            full_time_range=time_range,
            t_reference=t_reference,
            time_selection="touching",
            time_dim=records_points.kdims[0],
            peaks=peaks,
            show_largest=show_largest,
        ),
        streams=[time_stream],
    )

    layout = time_v_channel + peak_wvs + array_plot["top"] + array_plot["bottom"]
    return layout.cols(2)


def hook(plot, x_range, debug=False):
    """Hook to set the same RangeX stream for event display and records matrix, voodoo....

    Note:
        Works only in the following order:

        1. Create holoviews
        2. hv.render plot
        3. set bokeh x_range as holoviews x_range

        Does not work first with bokeh and then with holoviews. Why?
        I have no clue....

    """
    if debug:
        print("Hook: ", x_range)
    if not x_range:
        _hook_get_xrange(plot, debug)
    else:
        _hook_set_xrange(plot, x_range, debug)


def _hook_get_xrange(plot, debug):
    straxen._BOKEH_X_RANGE = plot.handles["x_range"]
    if debug:
        print("Get", straxen._BOKEH_X_RANGE)


def _hook_set_xrange(plot, x_range, debug):
    if debug:
        print("Set", x_range)
    plot.state.x_range = x_range



import strax
import straxen

from straxen.plugins.defaults import MV_PREAMBLE, MV_HIT_DEFAULTS
from straxen.plugins.records_nv.records_nv import nVETOPulseProcessing


export, __all__ = strax.exporter()


@export
class muVETOPulseProcessing(nVETOPulseProcessing):
    __doc__ = MV_PREAMBLE + (nVETOPulseProcessing.__doc__ or "")
    __version__ = "0.0.1"

    depends_on = "raw_records_mv"
    provides = "records_mv"
    data_kind = "records_mv"
    child_plugin = True

    save_outside_hits_mv = straxen.URLConfig(
        default=MV_HIT_DEFAULTS["save_outside_hits_mv"],
        track=True,
        infer_type=False,
        child_option=True,
        parent_option_name="save_outside_hits_nv",
        help="Save (left, right) samples besides hits; cut the rest",
    )

    baseline_samples_mv = straxen.URLConfig(
        default=100,
        track=True,
        infer_type=False,
        child_option=True,
        parent_option_name="baseline_samples_nv",
        help="Number of samples to use at the start of the pulse to determine the baseline",
    )

    hit_min_amplitude_mv = straxen.URLConfig(
        infer_type=False,
        default=MV_HIT_DEFAULTS["hit_min_amplitude_mv"],
        track=True,
        help=(
            "Minimum hit amplitude in ADC counts above baseline. "
            "Specify as a tuple of length n_mveto_pmts, or a number."
        ),
    )

    check_raw_record_overlaps = straxen.URLConfig(
        default=True,
        track=False,
        infer_type=False,
        help="Crash if any of the pulses in raw_records overlap with others in the same channel",
    )

    def setup(self):
        self.baseline_samples = self.baseline_samples_mv
        self.hit_thresholds = self.hit_min_amplitude_mv

    def infer_dtype(self):
        record_length = strax.record_length_from_dtype(
            self.deps["raw_records_mv"].dtype_for("raw_records_mv")
        )
        dtype = strax.record_dtype(record_length)
        return dtype

    def compute(self, raw_records_mv):
        if self.check_raw_record_overlaps:
            straxen.check_overlaps(raw_records_mv, n_channels=3000)
        return super().compute(raw_records_mv)



import strax
import straxen


def check_superruns_wrapper(func):
    """Check whether all configs for superruns-allowed plugins are the same for subruns."""

    def wrapper(self, run_id, targets=tuple(), **kwargs):
        is_superrun = run_id.startswith("_")
        if (
            not is_superrun
            or not self.context_config["check_superrun_configs"]
            or kwargs.get("combining", False)
        ):
            return func(self, run_id=run_id, targets=targets, **kwargs)

        configs = self.time_dependent_configs(run_id, targets)
        if not configs:
            return func(self, run_id=run_id, targets=targets, **kwargs)
        superrun_hashes = {run_id: self.hashed_url_configs(configs)}

        # collect all url config of all subruns and convert them into hash
        sub_run_spec = self.run_metadata(run_id, projection="sub_run_spec")["sub_run_spec"]
        for sub_run_id in sub_run_spec:
            configs = self.time_dependent_configs(sub_run_id, targets)
            superrun_hashes[sub_run_id] = self.hashed_url_configs(configs)

        # check whether all subruns have the same configs
        seen = strax.deterministic_hash(superrun_hashes[run_id])
        for sub_run_id, value in superrun_hashes.items():
            if seen != strax.deterministic_hash(value):
                diff = [k for k in value if value[k] != superrun_hashes[run_id][k]]
                raise ValueError(
                    "Superrun configs are not the same for all subruns. "
                    f"Specifically, the following configs are different: {diff}."
                )

        return func(self, run_id=run_id, targets=targets, **kwargs)

    return wrapper


# Manually decorate context class
_strax_context_check_superruns_decorated = True
if "check_superrun_configs" not in strax.Context.takes_config:
    strax.Context = strax.takes_config(
        strax.Option(
            name="check_superrun_configs",
            default=True,
            type=bool,
            help="If True, check whether all subruns' config are the same.",
        ),
    )(strax.Context)
    _strax_context_check_superruns_decorated = False
if "plugin_attr_convert" not in strax.Context.takes_config:
    strax.Context = strax.takes_config(
        strax.Option(
            name="plugin_attr_convert",
            default=("run_id", "algorithm", "sr"),
            type=(list, tuple),
            help="The attributes that should be get from the plugin.",
        ),
    )(strax.Context)
    _strax_context_check_superruns_decorated = False
if not _strax_context_check_superruns_decorated:
    # Overwrite get_components method
    strax.Context.get_components = check_superruns_wrapper(strax.Context.get_components)


@strax.Context.add_method
def time_dependent_configs(self, run_id, targets, superrun_only=True):
    targets = strax.to_str_tuple(targets)
    plugins = self._get_plugins(targets=targets, run_id=run_id)
    configs = dict()
    for data_type, plugin in plugins.items():
        if not superrun_only or (plugin.allow_superrun and superrun_only):
            for k, v in plugin.config.items():
                if not isinstance(v, str):
                    continue
                configs[k] = (v, plugins[data_type])
    return configs


@strax.Context.add_method
def hashed_url_configs(self, configs):
    """Convert all url into string and hash them."""
    # get all configs for superruns-allowed plugins if they are str
    hash = dict()
    for key, value in configs.items():
        url, plugin = value
        url = straxen.URLConfig.format_url_kwargs(url)
        arg, extra_kwargs = straxen.URLConfig.split_url_kwargs(url)
        # only extract value, avoid attr=map
        if "attr" in extra_kwargs:
            extra_kwargs["attr"] = "value"
        # transform run_id and algorithm from plugin
        # run_id and algorithm are the only two keys that
        # can be extracted from plugin for now
        # if later we have more keys need to be converted from plugin
        # to make xedocs work, we need to add them here
        for k in self.context_config["plugin_attr_convert"]:
            if k in extra_kwargs and straxen.URLConfig.NAMESPACE_SEP in extra_kwargs[k]:
                extra_kwargs[k] = getattr(
                    plugin,
                    extra_kwargs[k].partition(straxen.URLConfig.NAMESPACE_SEP)[-1],
                )
        # filter out plugin related kwargs
        _extra_kwargs = dict()
        for k, v in extra_kwargs.items():
            flag = not isinstance(v, str)
            flag |= isinstance(v, str) and straxen.URLConfig.PLUGIN_ATTR_PREFIX not in v
            if flag:
                _extra_kwargs[k] = v
        url = straxen.URLConfig.format_url_kwargs(arg, **_extra_kwargs)
        # replace all possible missing attr like in runstart://plugin.run_id
        # the remaining plugin related attr have not been replaced because they are not kwargs
        for k in self.context_config["plugin_attr_convert"]:
            if f"plugin.{k}" in url:
                url = url.replace(f"plugin.{k}", getattr(plugin, k))
        if "resource" in url:
            protocol, arg, kwargs = straxen.URLConfig.url_to_ast(url)
            while protocol != "resource":
                protocol, arg, kwargs = arg
            if isinstance(arg, tuple):
                url = straxen.URLConfig.ast_to_url(arg)
            else:
                url = arg
        evaluated = straxen.URLConfig.evaluate_dry(url)
        try:
            strax.hashablize(evaluated)
        except TypeError as e:
            raise ValueError(
                f"Cannot hash url {url} converted from config {key} in superrun safeguard. "
                f"Because its evaluated value not hashable: {evaluated}. "
                f"Error: {e}"
            )
        hash[key] = strax.deterministic_hash(evaluated)
    return hash



import os
import json
import pytz
import typing
import fsspec
import runpy
import tarfile
import tempfile
import shutil
from typing import Container, Iterable
import numpy as np
from scipy.interpolate import interp1d
from immutabledict import immutabledict

from utilix import xent_collection
import utilix
import strax
import straxen
from .url_config import URLConfig


def get_item_or_attr(obj, key, default=None):
    if isinstance(obj, dict):
        return obj.get(key, default)
    return getattr(obj, key, default)


@URLConfig.register("resource")
def get_resource(name: str, fmt: str = "text", readable: bool = False, **kwargs):
    """Fetch a straxen resource, allow a direct download using <fmt='abs_path'> otherwise kwargs are
    passed directly to straxen.get_resource."""
    if fmt == "abs_path":
        downloader = utilix.mongo_storage.MongoDownloader()
        return downloader.download_single(name, human_readable_file_name=readable)
    return straxen.get_resource(name, fmt=fmt, readable=readable)


@URLConfig.register("fsspec")
def read_file(path: str, **kwargs):
    """Support fetching files from arbitrary filesystems."""
    with fsspec.open(path, **kwargs) as f:
        content = f.read()
    return content


@URLConfig.register("json")
def read_json(content: str, **kwargs):
    """Load json string as a python object."""
    return json.loads(content)


@URLConfig.register("take")
def get_key(container: Container, take=None, **kwargs):
    """Return a single element of a container."""
    if not isinstance(container, dict):
        raise ValueError(f"Container is not a dict but a {type(container)}")
    if take is None:
        return container
    if not isinstance(take, list):
        take = [take]

    # support for multiple keys for
    # nested objects
    for t in take:
        container = container[t]  # type: ignore

    return container


@URLConfig.register("format")
def format_arg(arg: str, **kwargs):
    """Apply pythons builtin format function to a string."""
    return arg.format(**kwargs)


@URLConfig.register("itp_map")
def load_map(some_map, method="WeightedNearestNeighbors", scale_coordinates=None, **kwargs):
    """Make an InterpolatingMap."""
    itp_map = straxen.InterpolatingMap(some_map, method=method, **kwargs)
    if scale_coordinates is not None:
        itp_map.scale_coordinates(scale_coordinates)
    return itp_map


@URLConfig.register("bodega")
def load_value(name: str, bodega_version=None):
    """Load a number from BODEGA file."""
    if bodega_version is None:
        raise ValueError("Provide version see e.g. tests/test_url_config.py")
    nt_numbers = straxen.get_resource("XENONnT_numbers.json", fmt="json")
    return nt_numbers[name][bodega_version]["value"]


@URLConfig.register("tf")
def open_neural_net(model_path: str, custom_objects=None, register=False, **kwargs):
    """Load a keras model from a keras file.

    If the model is a tar.gz file, it will be extracted and the registration.py file will be
    imported. This file should contain the registration of custom objects and the model should be
    saved as a .keras file.

    """
    # Nested import to reduce loading time of import straxen and it not
    # base requirement
    import tensorflow as tf

    if not os.path.exists(model_path):
        raise FileNotFoundError(f"No file at {model_path}")

    if register:
        with tempfile.TemporaryDirectory() as tmpdirname:
            tar = tarfile.open(model_path, mode="r:gz")
            tar.extractall(path=tmpdirname)
            # Execute the registration.py file
            # This file should contain the registration of custom objects
            # and the model should be saved as a .keras file
            # Assign the module to a unique name to avoid conflicts
            runpy.run_path(os.path.join(tmpdirname, "registration.py"))
            for filename in os.listdir(tmpdirname):
                if filename.endswith(".keras"):
                    return tf.keras.models.load_model(
                        os.path.join(tmpdirname, filename),
                        custom_objects=custom_objects,
                    )
        raise FileNotFoundError(f"No .keras file found in {model_path}!")
    else:
        if model_path.endswith(".keras"):
            return tf.keras.models.load_model(model_path, custom_objects=custom_objects)
        else:
            # If the model is not a .keras file,
            # copy it to a temporary directory as .keras file and load it
            with tempfile.TemporaryDirectory() as tmpdirname:
                filename = os.path.join(tmpdirname, os.path.basename(model_path) + ".keras")
                shutil.copy(model_path, filename)
                return tf.keras.models.load_model(filename, custom_objects=custom_objects)


@URLConfig.register("itp_dict")
def get_itp_dict(
    loaded_json, run_id=None, time_key="time", itp_keys="correction", **kwargs
) -> typing.Union[np.ndarray, typing.Dict[str, np.ndarray]]:
    """Interpolate a dictionary at the start time that is queried from a run-id.

    :param loaded_json: a dictionary with a time-series
    :param run_id: run_id
    :param time_key: key that gives the timestamps
    :param itp_keys: which keys from the dict to read. Should be comma (',') separated!
    :return: Interpolated values of dict at the start time, either returned as an np.ndarray (single
        value) or as a dict (multiple itp_dict_keys)

    """
    keys = strax.to_str_tuple(itp_keys.split(","))
    for key in list(keys) + [time_key]:
        if key not in loaded_json:
            raise KeyError(
                f"The json does contain the key '{key}'. Try one of: {loaded_json.keys()}"
            )

    times = loaded_json[time_key]

    # get start time of this run. Need to make tz-aware
    start = xent_collection().find_one({"number": int(run_id)}, {"start": 1})["start"]
    start = pytz.utc.localize(start).timestamp() * 1e9

    try:
        if len(strax.to_str_tuple(keys)) > 1:
            return {
                key: interp1d(times, loaded_json[key], bounds_error=True)(start) for key in keys
            }

        else:
            interp = interp1d(times, loaded_json[keys[0]], bounds_error=True)
            return interp(start)
    except ValueError as e:
        raise ValueError(f"Correction is not defined for run {run_id}") from e


@URLConfig.register("rekey_dict")
def rekey_dict(d, replace_keys="", with_keys=""):
    """Replace the keys of a dictionary.

    :param d: dictionary that will have its keys renamed
    :param replace_keys: comma-separated string of keys that will be replaced
    :param with_keys: comma-separated string of keys that will replace the replace_keys
    :return: dictionary with renamed keys

    """
    new_dict = d.copy()
    replace_keys = strax.to_str_tuple(replace_keys.split(","))
    with_keys = strax.to_str_tuple(with_keys.split(","))
    if len(replace_keys) != len(with_keys):
        raise RuntimeError("replace_keys and with_keys must have the same length")
    for old_key, new_key in zip(replace_keys, with_keys):
        new_dict[new_key] = new_dict.pop(old_key)
    return new_dict


@URLConfig.register("objects-to-dict")
def objects_to_dict(objects: list, key_attr=None, value_attr="value", immutable=False):
    """Converts a list of objects/dicts to a single dictionary by taking the key and value from each
    of the objects/dicts. If key_attr is not provided, the list index is used as the key.

    :param objects: list of objects/dicts that will be converted to a dictionary
    :param key_attr: key/attribute of the objects that will be used as key in the dictionary
    :param value_attr: key/attribute of the objects that will be used as value in the dictionary

    """
    if not isinstance(objects, Iterable):
        raise TypeError(
            f"The objects-to-dict protocol expects an iterable "
            f"of objects but received {type(objects)} instead."
        )
    result = {}
    for i, obj in enumerate(objects):
        key = i if key_attr is None else get_item_or_attr(obj, key_attr)
        result[key] = get_item_or_attr(obj, value_attr)

    if immutable:
        result = immutabledict(result)

    return result


@URLConfig.register("list-to-array")
def objects_to_array(objects: list):
    """Converts a list of objects/dicts to a numpy array.

    :param objects: Any list of objects

    """

    if not isinstance(objects, Iterable):
        raise TypeError(
            f"The list-to-array protocol expects an "
            f"iterable but recieved a {type(objects)} instead"
        )

    return np.array(objects)


@URLConfig.register("run_doc")
def read_rundoc(path, run_id=None, default=None):
    """Read a path from the json rundoc metadata.

    :param path: keys of json rundoc metada.
        e.g. `comments` for reading comments for a specific run

    """
    if run_id is None:
        raise ValueError("rundoc protocol: missing run_id.")
    runs = xent_collection()
    rundoc = runs.find_one({"number": int(run_id)}, {"_id": 0, path: 1})
    if rundoc is None:
        raise ValueError(f"No rundoc found for run {run_id}")

    for part in path.split("."):
        if isinstance(rundoc, list) and part.isdigit() and len(rundoc) > int(part):
            rundoc = rundoc[int(part)]
        elif isinstance(rundoc, dict) and part in rundoc:
            rundoc = rundoc[part]
        elif default is not None:
            return default
        else:
            raise ValueError(f"No path {path} found in rundoc for run {run_id}")
    return rundoc


@URLConfig.register("pad-array")
def get_paded_array(arr: np.ndarray, pad_value=0, pad_left=0, pad_right=0):
    """Pad the array with pad_value on the left and right side."""
    return np.pad(arr, (pad_left, pad_right), constant_values=pad_value)


@URLConfig.register("jax")
def open_jax_model(model_path: str, **kwargs):
    """Open and deserialize a JAX model from a tar file.

    This function is registered with straxen's URLConfig under the 'jax' key.
    It opens a tar file containing serialized JAX models, extracts the requested
    model based on the provided parameters, and returns the deserialized model.

    Args:
        model_path (str): Path to the tar file containing the JAX models.
        **kwargs: Additional keyword arguments.
            Required:
                n_poly (int): Number of polynomials used in the model.
                sig (float): Sigma value used in the model.

    Returns:
        callable: The deserialized JAX model as a callable function.

    Raises:
        FileNotFoundError: If the specified model_path does not exist.
        ValueError: If the requested model is not found in the tar file.

    Note:
        This function uses a nested import of 'jax.export' to reduce the loading
        time of importing straxen, as JAX is not a base requirement.

    """
    # Nested import to reduce loading time of import straxen as it's not
    # a base requirement
    from jax import export

    # Check if the model file exists
    if not os.path.exists(model_path):
        raise FileNotFoundError(f"No file at {model_path}")

    # Extract required parameters from kwargs
    n_poly = kwargs["n_poly"]
    sig = kwargs["sig"]

    # Open the tar file and extract the requested model
    with tarfile.open(model_path, "r") as f:
        names = f.getnames()

        # Construct the filename based on n_poly and sig
        filename = f"{int(n_poly):02d}_{float(sig) * 1000:3.0f}"

        # Check if the requested model exists in the tar file
        if filename not in names:
            raise ValueError("Requested model not in tarfile!")

        # Extract and read the serialized JAX object
        file_obj = f.extractfile(filename)
        assert file_obj is not None
        serialized_jax_object = file_obj.read()
    # Deserialize the JAX object and return its callable function
    return export.deserialize(serialized_jax_object).call


@URLConfig.register("runstart")
def get_run_start(run_id):
    """Protocol which returns start time of a given run as unix time in ns."""
    rundb = utilix.xent_collection()
    doc = rundb.find_one(
        {"number": int(run_id)},
        projection={
            "start": 1,
        },
    )
    start_time = doc["start"]
    start_time_unix = start_time.replace(tzinfo=pytz.utc).timestamp()
    start_time_unix = np.int64(start_time_unix) * straxen.units.s
    return start_time_unix



from strax import MongoFrontend, exporter
from straxen import uconfig

export, __all__ = exporter()

default_online_collection = "online_monitor"


@export
def get_mongo_uri(
    user_key="pymongo_user", pwd_key="pymongo_password", url_key="pymongo_url", header="RunDB"
):
    user = uconfig.get(header, user_key)
    pwd = uconfig.get(header, pwd_key)
    url = uconfig.get(header, url_key)
    return f"mongodb://{user}:{pwd}@{url}"


@export
class OnlineMonitor(MongoFrontend):
    """Online monitor Frontend for Saving data temporarily to the database."""

    def __init__(
        self,
        uri=None,
        take_only=None,
        database=None,
        col_name=default_online_collection,
        readonly=True,
        *args,
        **kwargs,
    ):
        if take_only is None:
            raise ValueError(
                f"Specify which data_types to accept! Otherwise the DataBase will be overloaded"
            )
        if uri is None and readonly:
            uri = get_mongo_uri()
        elif uri is None and not readonly:
            # 'not readonly' means that you want to write. Let's get
            # your admin credentials:
            uri = get_mongo_uri(
                header="rundb_admin",
                user_key="mongo_rdb_username",
                pwd_key="mongo_rdb_password",
                url_key="mongo_rdb_url",
            )

        if database is None:
            database = uconfig.get("RunDB", "pymongo_database")

        super().__init__(
            uri=uri, database=database, take_only=take_only, col_name=col_name, *args, **kwargs
        )
        self.readonly = readonly



"""Small utility to print versions of software.

Example:
    straxen_print_versions strax straxen cutax wfsim pema

"""

import argparse


def main():
    parser = argparse.ArgumentParser(
        description="Print software versions using straxen.print_versions",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "packages",
        metavar="PACKAGES",
        default=["strax", "straxen", "cutax"],
        nargs="*",
        help="Which packages to print installations of",
    )
    args = parser.parse_args()

    from strax import to_str_tuple
    from straxen import print_versions

    print_versions(to_str_tuple(args.packages))


if __name__ == "__main__":
    main()



from straxen.plugins.events.corrected_areas import CorrectedAreas
import strax
import numpy as np
import straxen

export, __all__ = strax.exporter()


@export
class PeakCorrectedAreas(CorrectedAreas):
    """Pluging to apply corrections on peak level assuming that the main S1 is the only physical
    S1."""

    __version__ = "0.0.1"

    depends_on = ("peak_basics", "peak_positions", "peak_per_event")
    data_kind = "peaks"
    provides = "peak_corrections"

    electron_drift_velocity = straxen.URLConfig(
        default="xedocs://electron_drift_velocities?attr=value&run_id=plugin.run_id&version=ONLINE",
        cache=True,
        help="Vertical electron drift velocity in cm/ns (1e4 m/ms)",
    )

    electron_drift_time_gate = straxen.URLConfig(
        default="xedocs://electron_drift_time_gates?attr=value&run_id=plugin.run_id&version=ONLINE",
        help="Electron drift time from the gate in ns",
        cache=True,
    )

    def infer_dtype(self):
        dtype = strax.time_fields + [
            (
                (
                    ("Bias-corrected S1 area before xyz correction [PE]"),
                    "cs1_wo_xyzcorr",
                ),
                np.float32,
            ),
            (
                (
                    ("Bias-corrected S2 area before xy correction [PE]"),
                    "cs2_wo_xycorr",
                ),
                np.float32,
            ),
            (
                (
                    (
                        "Corrected area of S2 before elife correction "
                        "(s2 xy correction + SEG/EE correction applied) [PE]"
                    ),
                    "cs2_wo_elifecorr",
                ),
                np.float32,
            ),
            (
                (
                    (
                        "Corrected area of S2 before SEG/EE and elife corrections "
                        "(s2 xy correction applied) [PE]"
                    ),
                    "cs2_wo_timecorr",
                ),
                np.float32,
            ),
            (
                (
                    "Fraction of area seen by the top PMT array for corrected S2",
                    "cs2_area_fraction_top",
                ),
                np.float32,
            ),
            (("Corrected area of S2 in the bottom PMT array [PE]", "cs2_bottom"), np.float32),
            (("Corrected area of S2 [PE]", "cs2"), np.float32),
            (
                (
                    "Correction factor for the S1 area based on S2 position",
                    "s1_xyz_correction_factor",
                ),
                np.float32,
            ),
            (
                (
                    "Relative light yield correction factor for the S1 area",
                    "s1_rel_light_yield_correction_factor",
                ),
                np.float32,
            ),
            (("z position of the multiscatter peak", "z_obs_ms"), np.float32),
        ]
        return dtype

    def compute(self, peaks):
        result = np.zeros(len(peaks), self.dtype)
        result["time"] = peaks["time"]
        result["endtime"] = peaks["endtime"]

        # Get z position of the peak
        z_obs = -self.electron_drift_velocity * peaks["drift_time"]
        z_obs = z_obs + self.electron_drift_velocity * self.electron_drift_time_gate
        result["z_obs_ms"] = z_obs

        # S1 correction factors
        s1_mask = peaks["type"] == 1
        if np.any(s1_mask):
            s1_bias = self.s1_bias_map(peaks["area"][s1_mask].reshape(-1, 1))
            cs1_wo_xyzcorr = peaks["area"][s1_mask] / (1 + s1_bias.flatten())
            result["cs1_wo_xyzcorr"][s1_mask] = cs1_wo_xyzcorr

        # S2 correction factors
        s2_mask = peaks["type"] == 2
        if np.any(s2_mask):
            s2_bias = self.s2_bias_map(peaks["area"][s2_mask].reshape(-1, 1))
            cs2_wo_xycorr = peaks["area"][s2_mask] / (1 + s2_bias.flatten())
            result["cs2_wo_xycorr"][s2_mask] = cs2_wo_xycorr

        peak_positions = np.vstack([peaks["x"], peaks["y"], z_obs]).T
        result["s1_xyz_correction_factor"] = 1 / self.s1_xyz_map(peak_positions)
        result["s1_rel_light_yield_correction_factor"] = 1 / self.rel_light_yield

        # s2 corrections
        s2_top_map_name, s2_bottom_map_name = self.s2_map_names()

        seg, avg_seg, ee = self.seg_ee_correction_preparation()

        # now can start doing corrections

        # S2(x,y) corrections use the observed S2 positions
        s2_positions = np.vstack([peaks["x"], peaks["y"]]).T

        # corrected s2 with s2 xy map only, i.e. no elife correction
        # this is for s2-only events which don't have drift time info

        cs2_top_xycorr = (
            peaks["area"]
            * peaks["area_fraction_top"]
            / self.s2_xy_map(s2_positions, map_name=s2_top_map_name)
        )
        cs2_bottom_xycorr = (
            peaks["area"]
            * (1 - peaks["area_fraction_top"])
            / self.s2_xy_map(s2_positions, map_name=s2_bottom_map_name)
        )

        # For electron lifetime corrections to the S2s,
        # use drift time computed using the main S1.

        elife_correction = np.exp(peaks["drift_time"] / self.elife)
        result["cs2_wo_timecorr"] = (cs2_top_xycorr + cs2_bottom_xycorr) * elife_correction

        for partition, func in self.regions.items():
            # partitioned SE and EE
            partition_mask = func(peaks["x"], peaks["y"])

            # Correct for SEgain and extraction efficiency
            seg_ee_corr = seg[partition] / avg_seg[partition] * ee[partition]

            # note that these are already masked!
            cs2_top_wo_elifecorr = cs2_top_xycorr[partition_mask] / seg_ee_corr
            cs2_bottom_wo_elifecorr = cs2_bottom_xycorr[partition_mask] / seg_ee_corr

            result["cs2_wo_elifecorr"][partition_mask] = (
                cs2_top_wo_elifecorr + cs2_bottom_wo_elifecorr
            )

            # cs2aft doesn't need elife/time corrections as they cancel
            result["cs2_area_fraction_top"][partition_mask] = cs2_top_wo_elifecorr / (
                cs2_top_wo_elifecorr + cs2_bottom_wo_elifecorr
            )

            result["cs2"][partition_mask] = (
                result["cs2_wo_elifecorr"][partition_mask] * elife_correction[partition_mask]
            )
            result["cs2_bottom"][partition_mask] = (
                cs2_bottom_wo_elifecorr * elife_correction[partition_mask]
            )

        not_s2_mask = peaks["type"] != 2
        result["cs2_wo_timecorr"][not_s2_mask] = np.nan
        result["cs2_wo_elifecorr"][not_s2_mask] = np.nan
        result["cs2_area_fraction_top"][not_s2_mask] = np.nan
        result["cs2"][not_s2_mask] = np.nan
        result["z_obs_ms"][not_s2_mask] = np.nan
        result["cs2_bottom"][not_s2_mask] = np.nan
        result["s1_xyz_correction_factor"][not_s2_mask] = np.nan
        result["s1_rel_light_yield_correction_factor"][not_s2_mask] = np.nan
        return result



import numba
import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class LEDAfterpulseProcessing(strax.Plugin):
    """Plugin for processing LED afterpulses.

    Detect LED pulses and afterpulses (APs) in raw_records waveforms. Compute the AP datatype.

    """

    __version__ = "0.6.0"
    depends_on = "raw_records"
    data_kind = "afterpulses"
    provides = "afterpulses"
    compressor = "zstd"
    parallel = "process"
    rechunk_on_save = True

    gain_model = straxen.URLConfig(
        infer_type=False,
        help="PMT gain model. Specify as (model_type, model_config)",
    )

    n_tpc_pmts = straxen.URLConfig(
        type=int,
        help="Number of PMTs in TPC",
    )

    LED_hit_left_boundary = straxen.URLConfig(
        default=50,
        infer_type=False,
        help="Left boundary after which the first hit marks the position of the LED window",
    )

    LED_hit_right_boundary = straxen.URLConfig(
        default=110,
        infer_type=False,
        help="Right boundary after which the LED hit will no longer be searched",
    )

    LED_window_width = straxen.URLConfig(
        default=20,
        infer_type=False,
        help="Width of the window in which hits are merged into the LED hit after the first hit",
    )

    baseline_samples = straxen.URLConfig(
        default=40,
        infer_type=False,
        help="Number of samples to use at start of WF to determine the baseline",
    )

    hit_min_amplitude = straxen.URLConfig(
        track=True,
        infer_type=False,
        default=(
            "list-to-array://"
            "xedocs://hit_thresholds"
            "?as_list=True"
            "&sort=pmt"
            "&attr=value"
            "&detector=tpc"
            "&run_id=plugin.run_id"
            "&version=ONLINE"
        ),
        help="Minimum hit amplitude in ADC counts above baseline. "
        "Specify as a tuple of length n_tpc_pmts, or a number,"
        'or a string like "legacy-thresholds://pmt_commissioning_initial" which means calling'
        "hitfinder_thresholds.py"
        'or url string like "xedocs://hit_thresholds_tpc?version=ONLINE" which means'
        "calling xedocs.",
    )

    hit_min_height_over_noise = straxen.URLConfig(
        default=4,
        infer_type=False,
        help=(
            "Minimum hit amplitude in numbers of baseline_rms above baseline."
            "Actual threshold used is max(hit_min_amplitude, hit_min_"
            "height_over_noise * baseline_rms)."
        ),
    )

    save_outside_hits = straxen.URLConfig(
        default=(3, 20),
        infer_type=False,
        help="Save (left, right) samples besides hits; cut the rest",
    )

    def infer_dtype(self):
        dtype = dtype_afterpulses()
        return dtype

    def setup(self):
        self.to_pe = self.gain_model
        self.hit_thresholds = self.hit_min_amplitude
        self.hit_left_extension, self.hit_right_extension = self.save_outside_hits

    def compute(self, raw_records):
        # Convert everything to the records data type -- adds extra fields.
        records = strax.raw_to_records(raw_records)
        del raw_records

        # calculate baseline and baseline rms
        strax.baseline(records, baseline_samples=self.baseline_samples, flip=True)

        # find all hits
        hits = strax.find_hits(
            records,
            min_amplitude=self.hit_thresholds,
            min_height_over_noise=self.hit_min_height_over_noise,
        )

        # sort hits by record_i and time, then find LED hit and afterpulse
        # hits within the same record
        hits_ap = find_ap(
            hits,
            records,
            LED_hit_left_boundary=self.LED_hit_left_boundary,
            LED_hit_right_boundary=self.LED_hit_right_boundary,
            LED_window_width=self.LED_window_width,
            hit_left_extension=self.hit_left_extension,
            hit_right_extension=self.hit_right_extension,
        )

        hits_ap["area_pe"] = hits_ap["area"] * self.to_pe[hits_ap["channel"]]
        hits_ap["height_pe"] = hits_ap["height"] * self.to_pe[hits_ap["channel"]]

        return hits_ap


@export
def find_ap(
    hits,
    records,
    LED_hit_left_boundary,
    LED_hit_right_boundary,
    LED_window_width,
    hit_left_extension,
    hit_right_extension,
):
    """Find afterpulses (APs) in the given hits data within specified LED hit boundaries and
    extensions.

    Parameters
    ----------
    hits :
        Array containing hit data.
    records :
        Array containing record data.
    LED_hit_left_boundary :
        Left boundary of the LED hit window.
    LED_hit_right_boundary :
        Right boundary of the LED hit window.
    LED_window_width :
        Extension to the right of the first hit found in the LED hit window
        within which hits are merged into the LED hit.
    hit_left_extension :
        Extension to the left of the hit window.
    hit_right_extension :
        Extension to the right of the hit window.

    Returns
    -------
    Array containing afterpulse data.

    Notes
    -----
    - Hits to the left of the LED_hit_left_boundary are ignored.
    - If no hit is found between LED_hit_left_boundary and LED_hit_right_boundary
    the record is skipped.
    - The merged LED hits are also saved and can be selected for by having
    t_delay = 0 by definition.

    """

    buffer = np.zeros(len(hits), dtype=dtype_afterpulses())

    if not len(hits):
        return buffer

    # sort hits first by record_i, then by time
    hits_sorted = strax.stable_sort(hits, order=("record_i", "time"))
    res = _find_ap(
        hits_sorted,
        records,
        LED_hit_left_boundary,
        LED_hit_right_boundary,
        LED_window_width,
        hit_left_extension,
        hit_right_extension,
        buffer=buffer,
    )
    return res


@numba.jit(nopython=True, nogil=True, cache=True)
def _find_ap(
    hits,
    records,
    LED_hit_left_boundary,
    LED_hit_right_boundary,
    LED_window_width,
    hit_left_extension,
    hit_right_extension,
    buffer=None,
):
    # hits need to be sorted by record_i, then time!
    offset = 0

    is_LED = False
    t_LED = None  # The position of the total LED pulse.
    t_LED_window = None  # The position that defines the LED integration window.

    prev_record_i = hits[0]["record_i"]
    record_data = records[prev_record_i]["data"]
    record_len = records[prev_record_i]["length"]
    baseline_fpart = records[prev_record_i]["baseline"] % 1

    for h_i, h in enumerate(hits):
        if h["record_i"] > prev_record_i:
            # start of a new record
            is_LED = False
            # only increment buffer if the old one is not empty! this happens
            # when no (LED) hit is found in the previous record
            if not buffer[offset]["time"] == 0:
                offset += 1
            prev_record_i = h["record_i"]
            record_data = records[prev_record_i]["data"]
            baseline_fpart = records[prev_record_i]["baseline"] % 1

        res = buffer[offset]

        if h["left"] < LED_hit_left_boundary:
            # if hit is before LED hit window: discard
            continue

        if (h["left"] < LED_hit_right_boundary) and not is_LED:
            # this is the first hit in the LED hit window
            fill_hitpars(
                res,
                h,
                hit_left_extension,
                hit_right_extension,
                record_data,
                record_len,
                baseline_fpart,
            )

            t_LED_window = res[
                "sample_10pc_area"
            ]  # Set it to the first hit in the LED hit boundary.
            t_LED = res["sample_10pc_area"]
            is_LED = True

            continue

        if not is_LED:
            # No hit found in the LED hit window: skip to next record
            continue

        if h["left"] < (t_LED_window + LED_window_width):
            # This hit is still inside the LED window: extend the LED hit
            fill_hitpars(
                res,
                h,
                hit_left_extension,
                hit_right_extension,
                record_data,
                record_len,
                baseline_fpart,
                extend=True,
            )

            # set the LED time in the current WF
            t_LED = res["sample_10pc_area"]

            continue

        # Here begins a new hit after the LED window

        # if a hit is completely inside the previous hit's right_extension,
        # then skip it (because it's already included in the previous hit)
        if h["right"] <= res["right_integration"]:
            continue

        # if a hit only partly overlaps with the previous hit's right_
        # extension, merge them (extend previous hit by this one)
        if h["left"] <= res["right_integration"]:
            fill_hitpars(
                res,
                h,
                hit_left_extension,
                hit_right_extension,
                record_data,
                record_len,
                baseline_fpart,
                extend=True,
            )

            res["tdelay"] = res["sample_10pc_area"] - t_LED

            continue

        # an actual new hit increases the buffer index
        offset += 1
        res = buffer[offset]

        fill_hitpars(
            res,
            h,
            hit_left_extension,
            hit_right_extension,
            record_data,
            record_len,
            baseline_fpart,
        )

        res["tdelay"] = res["sample_10pc_area"] - t_LED

    return buffer[:offset]


@export
@numba.jit(nopython=True, nogil=True, cache=True)
def get_sample_area_quantile(data, quantile, baseline_fpart):
    """Return the index of the first sample in the hit where the integrated area of the hit to that
    index is above the specified quantile of the total area.

    Parameters:
    - data :
        Array containing the baselined waveform data of the hit.
    - quantile :
        The quantile (0 to 1) representing the threshold for the area.
    - baseline_fpart :
        Fractional part of the baseline (baseline % 1) of the record

    Return:
    - int: The index of the first sample where the area exceeds the quantile of the total area.
           If no such sample is found, returns 0.

    Notes:
    - If no quantile is found where the area exceeds the threshold, it returns 0. This is
    usually caused by real events in the baseline window, which can result in a negative
    area.

    """

    area = 0
    area_tot = data.sum() + len(data) * baseline_fpart

    for d_i, d in enumerate(data):
        area += d + baseline_fpart
        if area > (quantile * area_tot):
            return d_i
        if d_i == len(data) - 1:
            # if no quantile was found, something went wrong
            # (negative area due to wrong baseline, caused by real events that
            # by coincidence fall in the first samples of the trigger window)
            # print('no quantile found: set to 0')
            return 0

    # What happened here?!
    return 0


@numba.jit(nopython=True, nogil=True, cache=True)
def fill_hitpars(
    result,
    hit,
    hit_left_extension,
    hit_right_extension,
    record_data,
    record_len,
    baseline_fpart,
    extend=False,
):
    if not extend:  # fill first time only
        result["time"] = hit["time"] - hit_left_extension * hit["dt"]
        result["dt"] = hit["dt"]
        result["channel"] = hit["channel"]
        result["left"] = hit["left"]
        result["record_i"] = hit["record_i"]
        result["threshold"] = hit["threshold"]
        result["left_integration"] = hit["left"] - hit_left_extension
        result["height"] = hit["height"]

    # fill always (if hits are merged, only these will be updated)
    result["right"] = hit["right"]
    result["right_integration"] = hit["right"] + hit_right_extension
    if result["right_integration"] > record_len:
        result["right_integration"] = record_len  # cap right_integration at end of record
    result["length"] = result["right_integration"] - result["left_integration"]

    hit_data = record_data[result["left_integration"] : result["right_integration"]]
    result["area"] = hit_data.sum() + result["length"] * baseline_fpart
    result["sample_10pc_area"] = result["left_integration"] + get_sample_area_quantile(
        hit_data, 0.1, baseline_fpart
    )
    result["sample_50pc_area"] = result["left_integration"] + get_sample_area_quantile(
        hit_data, 0.5, baseline_fpart
    )
    if len(hit_data):
        result["max"] = result["left_integration"] + hit_data.argmax()

    if extend:  # only when merging hits
        result["height"] = max(result["height"], hit["height"])


@export
def dtype_afterpulses():
    """The afterpulse datatype.

    Return:
    - The afterpulse datatype

    """
    dtype_ap = strax.interval_dtype + [
        (("Integral in ADC x samples", "area"), np.int32),
        (("Pulse area in PE", "area_pe"), np.float32),
        (("Sample index in which hit starts", "left"), np.int16),
        (
            (
                "Sample index in which hit area succeeds 10% of total area",
                "sample_10pc_area",
            ),
            np.int16,
        ),
        (
            (
                "Sample index in which hit area succeeds 50% of total area",
                "sample_50pc_area",
            ),
            np.int16,
        ),
        (("Sample index of hit maximum", "max"), np.int16),
        (
            (
                "Index of first sample in record just beyond hit (exclusive bound)",
                "right",
            ),
            np.int16,
        ),
        (("Height of hit in ADC counts", "height"), np.int32),
        (("Height of hit in PE", "height_pe"), np.float32),
        (("Delay of hit w.r.t. LED hit in same WF, in samples", "tdelay"), np.int16),
        (
            (
                "Internal (temporary) index of fragment in which hit was found",
                "record_i",
            ),
            np.int32,
        ),
        (
            ("Index of sample in record where integration starts", "left_integration"),
            np.int16,
        ),
        (
            ("Index of first sample beyond integration region", "right_integration"),
            np.int16,
        ),
        (("ADC threshold applied in order to find hits", "threshold"), np.float32),
    ]
    return dtype_ap



from .config_tests import check_urls
from .url_config import URLConfig


@URLConfig.preprocessor
def sort_url_kwargs(url: str):
    """Reorders queries for urlconfigs to avoid hashing issues."""

    if isinstance(url, str) and URLConfig.SCHEME_SEP in url:
        return URLConfig.format_url_kwargs(url)
    return url


@URLConfig.preprocessor
def check_url_with_dispatcher(url: str):
    # Classes regex dispatcher to check url expressions
    if isinstance(url, str):
        check_urls(url)



import numpy as np
import strax

export, __all__ = strax.exporter()


@export
class EventShadow(strax.Plugin):
    """This plugin can calculate shadow for main S1 and main S2 in events. It also gives the
    position information of the previous peaks.

    References:
        * v0.1.4 reference: xenon:xenonnt:ac:prediction:shadow_ambience

    """

    __version__ = "0.1.4"
    depends_on = ("event_basics", "peak_shadow")
    provides = "event_shadow"

    def infer_dtype(self):
        dtype = []
        for main_peak, main_peak_desc in zip(["s1_", "s2_"], ["main S1", "main S2"]):
            # previous S1 can only cast time shadow,
            # previous S2 can cast both time & position shadow
            for key in ["s1_time_shadow", "s2_time_shadow", "s2_position_shadow"]:
                type_str, tp_desc, _ = key.split("_")
                dtype.append(
                    (
                        (
                            (
                                f"Largest {tp_desc} shadow casting from previous {type_str} to"
                                f" {main_peak_desc} [PE/ns]"
                            ),
                            f"{main_peak}shadow_{key}",
                        ),
                        np.float32,
                    )
                )
                dtype.append(
                    (
                        (
                            (
                                f"Time difference from the previous {type_str} casting largest"
                                f" {tp_desc} shadow to {main_peak_desc} [ns]"
                            ),
                            f"{main_peak}dt_{key}",
                        ),
                        np.int64,
                    )
                )
                # Only previous S2 peaks have (x,y)
                if "s2" in key:
                    dtype.append(
                        (
                            (
                                (
                                    f"X of previous s2 peak casting largest {tp_desc} shadow on"
                                    f" {main_peak_desc} [cm]"
                                ),
                                f"{main_peak}x_{key}",
                            ),
                            np.float32,
                        )
                    )
                    dtype.append(
                        (
                            (
                                (
                                    f"Y of previous s2 peak casting largest {tp_desc} shadow on"
                                    f" {main_peak_desc} [cm]"
                                ),
                                f"{main_peak}y_{key}",
                            ),
                            np.float32,
                        )
                    )
                # Only time shadow gives the nearest large peak
                if "time" in key:
                    dtype.append(
                        (
                            (
                                (f"The nearest previous large {type_str} to {main_peak_desc} [ns]"),
                                f"{main_peak}nearest_{type_str}",
                            ),
                            np.float32,
                        )
                    )
                    dtype.append(
                        (
                            (
                                (
                                    f"Time difference from the nearest previous large {type_str} to"
                                    f" {main_peak_desc} [ns]"
                                ),
                                f"{main_peak}nearest_dt_{type_str}",
                            ),
                            np.int64,
                        )
                    )
            # Also record the PDF of HalfCauchy when calculating S2 position shadow
            dtype.append(
                (
                    (
                        f"PDF describing correlation between previous s2 and {main_peak_desc}",
                        f"{main_peak}pdf_s2_position_shadow",
                    ),
                    np.float32,
                )
            )
        dtype += strax.time_fields
        return dtype

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)
        result = np.zeros(len(events), self.dtype)

        strax.set_nan_defaults(result)

        # 1. Assign peaks features to main S1 and main S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            res_i = result[event_i]
            # Fetch the features of main S1 and main S2
            for idx, main_peak in zip([event["s1_index"], event["s2_index"]], ["s1_", "s2_"]):
                if idx >= 0:
                    for key in ["s1_time_shadow", "s2_time_shadow", "s2_position_shadow"]:
                        type_str = key.split("_")[0]
                        res_i[f"{main_peak}shadow_{key}"] = sp[f"shadow_{key}"][idx]
                        res_i[f"{main_peak}dt_{key}"] = sp[f"dt_{key}"][idx]
                        if "time" in key:
                            res_i[f"{main_peak}nearest_{type_str}"] = sp[f"nearest_{type_str}"][idx]
                            res_i[f"{main_peak}nearest_dt_{type_str}"] = sp[
                                f"nearest_dt_{type_str}"
                            ][idx]
                        if "s2" in key:
                            res_i[f"{main_peak}x_{key}"] = sp[f"x_{key}"][idx]
                            res_i[f"{main_peak}y_{key}"] = sp[f"y_{key}"][idx]
                    # Record the PDF of HalfCauchy
                    res_i[f"{main_peak}pdf_s2_position_shadow"] = sp["pdf_s2_position_shadow"][idx]

        # 2. Set time and endtime for events
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)
        return result



import strax
from straxen.plugins.events.event_basics_vanilla import EventBasicsVanilla
from straxen.plugins.peaklets.peaklet_classification_som import som_additional_fields

export, __all__ = strax.exporter()


@export
class EventBasicsSOM(EventBasicsVanilla):
    """Adds SOM fields for S1 and S2 peaks to event basics."""

    __version__ = "0.0.1"
    child_plugin = True

    def _set_dtype_requirements(self):
        # Properties to store for each peak (main and alternate S1 and S2)
        # Add here SOM types:
        super()._set_dtype_requirements()
        self.peak_properties += tuple(som_additional_fields)



import strax
import straxen

from straxen.plugins.events_nv.events_nv import veto_event_dtype, nVETOEvents

export, __all__ = strax.exporter()


@export
class muVETOEvents(nVETOEvents):
    """Plugin which computes the boundaries of veto events."""

    depends_on = "hitlets_mv"  # type: ignore
    provides = "events_mv"
    data_kind = "events_mv"

    compressor = "zstd"
    child_plugin = True

    __version__ = "0.0.1"
    events_seen = 0

    event_left_extension_mv = straxen.URLConfig(
        default=0,
        track=True,
        type=int,
        child_option=True,
        parent_option_name="event_left_extension_nv",
        help="Extends event window this many [ns] to the left.",
    )
    event_resolving_time_mv = straxen.URLConfig(
        default=300,
        track=True,
        type=int,
        child_option=True,
        parent_option_name="event_resolving_time_nv",
        help="Resolving time for window coincidence [ns].",
    )
    event_min_hits_mv = straxen.URLConfig(
        default=3,
        track=True,
        type=int,
        child_option=True,
        parent_option_name="event_min_hits_nv",
        help="Minimum number of fully confined hitlets to define an event.",
    )

    def infer_dtype(self):
        self.name_event_number = "event_number_mv"
        self.channel_range = self.channel_map["mv"]
        self.n_channel = (self.channel_range[1] - self.channel_range[0]) + 1
        return veto_event_dtype(self.name_event_number, self.n_channel)

    def get_window_size(self):
        return self.event_left_extension_mv + self.event_resolving_time_mv + 1

    def compute(self, hitlets_mv):
        return super().compute(hitlets_mv)



from typing import Tuple
from immutabledict import immutabledict
import numba
import numpy as np

import strax
import straxen

export, __all__ = strax.exporter()
__all__.extend(["NO_PULSE_COUNTS"])


@export
class PulseProcessing(strax.Plugin):
    """
    Split raw_records into:
     - (tpc) records
     - aqmon_records
     - pulse_counts

    For TPC records, apply basic processing:
        1. Flip, baseline, and integrate the waveform
        2. Apply software HE veto after high-energy peaks.
        3. Find hits, apply linear filter, and zero outside hits.

    pulse_counts holds some average information for the individual PMT
    channels for each chunk of raw_records. This includes e.g.
    number of recorded pulses, lone_pulses (pulses which do not
    overlap with any other pulse), or mean values of baseline and
    baseline rms channel.
    """

    __version__ = "0.2.3"

    parallel = "process"
    rechunk_on_save = immutabledict(records=False, veto_regions=True, pulse_counts=True)
    compressor = "zstd"

    depends_on = "raw_records"

    provides: Tuple[str, ...] = ("records", "veto_regions", "pulse_counts")
    data_kind = {k: k for k in provides}
    save_when = immutabledict(
        records=strax.SaveWhen.TARGET,
        veto_regions=strax.SaveWhen.ALWAYS,
        pulse_counts=strax.SaveWhen.ALWAYS,
    )

    hev_gain_model = straxen.URLConfig(
        default=None, infer_type=False, help="PMT gain model used in the software high-energy veto."
    )

    baseline_samples = straxen.URLConfig(
        default=40,
        infer_type=False,
        help="Number of samples to use at the start of the pulse to determine the baseline",
    )

    # Tail veto options
    tail_veto_threshold = straxen.URLConfig(
        default=0,
        infer_type=False,
        help="Minimum peakarea in PE to trigger tail veto.Set to None, 0 or False to disable veto.",
    )

    tail_veto_duration = straxen.URLConfig(
        default=int(3e6), infer_type=False, help="Time in ns to veto after large peaks"
    )

    tail_veto_resolution = straxen.URLConfig(
        default=int(1e3),
        infer_type=False,
        help="Time resolution in ns for pass-veto waveform summation",
    )

    tail_veto_pass_fraction = straxen.URLConfig(
        default=0.05, infer_type=False, help="Pass veto if maximum amplitude above max * fraction"
    )

    tail_veto_pass_extend = straxen.URLConfig(
        default=3,
        infer_type=False,
        help="Extend pass veto by this many samples (tail_veto_resolution!)",
    )

    max_veto_value = straxen.URLConfig(
        default=None,
        infer_type=False,
        help=(
            "Optionally pass a HE peak that exceeds this absolute area. "
            "(if performing a hard veto, can keep a few statistics.)"
        ),
    )

    # PMT pulse processing options
    pmt_pulse_filter = straxen.URLConfig(
        default=None, infer_type=False, help="Linear filter to apply to pulses, will be normalized."
    )

    save_outside_hits = straxen.URLConfig(
        default=(3, 20),
        infer_type=False,
        help="Save (left, right) samples besides hits; cut the rest",
    )

    n_tpc_pmts = straxen.URLConfig(type=int, help="Number of TPC PMTs")

    check_raw_record_overlaps = straxen.URLConfig(
        default=True,
        track=False,
        infer_type=False,
        help="Crash if any of the pulses in raw_records overlap with others in the same channel",
    )

    allow_sloppy_chunking = straxen.URLConfig(
        default=False,
        track=False,
        infer_type=False,
        help=(
            "Use a default baseline for incorrectly chunked fragments. "
            "This is a kludge for improperly converted XENON1T data."
        ),
    )

    hit_min_amplitude = straxen.URLConfig(
        track=True,
        infer_type=False,
        default=(
            "list-to-array://"
            "xedocs://hit_thresholds"
            "?as_list=True"
            "&sort=pmt"
            "&attr=value"
            "&detector=tpc"
            "&run_id=plugin.run_id"
            "&version=ONLINE"
        ),
        help="Minimum hit amplitude in ADC counts above baseline. "
        "Specify as a tuple of length n_tpc_pmts, or a number,"
        'or a string like "pmt_commissioning_initial" which means calling'
        "hitfinder_thresholds.py",
    )

    def infer_dtype(self):
        # Get record_length from the plugin making raw_records
        self.record_length = strax.record_length_from_dtype(
            self.deps["raw_records"].dtype_for("raw_records")
        )

        dtype = dict()
        for p in self.provides:
            if "records" in p:
                dtype[p] = strax.record_dtype(self.record_length)
        dtype["veto_regions"] = strax.hit_dtype
        dtype["pulse_counts"] = pulse_count_dtype(self.n_tpc_pmts)

        return dtype

    def setup(self):
        self.hev_enabled = self.hev_gain_model is not None and self.tail_veto_threshold
        if self.hev_enabled:
            self.to_pe = self.hev_gain_model
        self.hit_thresholds = self.hit_min_amplitude

    def compute(self, raw_records, start, end):
        if self.check_raw_record_overlaps:
            check_overlaps(raw_records, n_channels=3000)

        # Throw away any non-TPC records; this should only happen for XENON1T
        # converted data
        raw_records = raw_records[raw_records["channel"] < self.n_tpc_pmts]

        # Convert everything to the records data type -- adds extra fields.
        r = strax.raw_to_records(raw_records)
        del raw_records

        # Do not trust in DAQ + strax.baseline to leave the
        # out-of-bounds samples to zero.
        strax.zero_out_of_bounds(r)

        strax.baseline(
            r,
            baseline_samples=self.baseline_samples,
            allow_sloppy_chunking=self.allow_sloppy_chunking,
            flip=True,
        )

        strax.integrate(r)

        pulse_counts = count_pulses(r, self.n_tpc_pmts)
        pulse_counts["time"] = start
        pulse_counts["endtime"] = end

        if len(r) and self.hev_enabled:
            r, r_vetoed, veto_regions = software_he_veto(
                r,
                self.to_pe,
                end,
                area_threshold=self.tail_veto_threshold,
                veto_length=self.tail_veto_duration,
                veto_res=self.tail_veto_resolution,
                pass_veto_extend=self.tail_veto_pass_extend,
                pass_veto_fraction=self.tail_veto_pass_fraction,
                max_veto_value=self.max_veto_value,
            )

            # In the future, we'll probably want to sum the waveforms
            # inside the vetoed regions, so we can still save the "peaks".
            del r_vetoed

        else:
            veto_regions = np.zeros(0, dtype=strax.hit_dtype)

        if len(r):
            # Find hits
            # -- before filtering,since this messes with the with the S/N
            hits = strax.find_hits(r, min_amplitude=self.hit_thresholds)

            if self.pmt_pulse_filter:
                # Filter to concentrate the PMT pulses
                strax.filter_records(r, np.array(self.pmt_pulse_filter))

            le, re = self.save_outside_hits
            r = strax.cut_outside_hits(r, hits, left_extension=le, right_extension=re)

            # Probably overkill, but just to be sure...
            strax.zero_out_of_bounds(r)

        return dict(records=r, pulse_counts=pulse_counts, veto_regions=veto_regions)


##
# Software HE Veto
##


@export
def software_he_veto(
    records,
    to_pe,
    chunk_end,
    area_threshold=int(1e5),
    veto_length=int(3e6),
    veto_res=int(1e3),
    pass_veto_fraction=0.01,
    pass_veto_extend=3,
    max_veto_value=None,
):
    """Veto veto_length (time in ns) after peaks larger than area_threshold (in PE).

    Further large peaks inside the veto regions are still passed:
    We sum the waveform inside the veto region (with time resolution
    veto_res in ns) and pass regions within pass_veto_extend samples
    of samples with amplitude above pass_veto_fraction times the maximum.

    :return: (preserved records, vetoed records, veto intervals).

    :param records: PMT records
    :param to_pe: ADC to PE conversion factors for the channels in records.
    :param chunk_end: Endtime of chunk to set as maximum ceiling for the veto period
    :param area_threshold: Minimum peak area to trigger the veto.
    Note we use a much rougher clustering than in later processing.
    :param veto_length: Time in ns to veto after the peak
    :param veto_res: Resolution of the sum waveform inside the veto region.
    Do not make too large without increasing integer type in some strax
    dtypes...
    :param pass_veto_fraction: fraction of maximum sum waveform amplitude to
    trigger veto passing of further peaks
    :param pass_veto_extend: samples to extend (left and right) the pass veto
    regions.
    :param max_veto_value: if not None, pass peaks that exceed this area
    no matter what.

    """
    veto_res = int(veto_res)
    if veto_res > np.iinfo(np.int16).max:
        raise ValueError("Veto resolution does not fit 16-bit int")
    veto_length = np.ceil(veto_length / veto_res).astype(np.int64) * veto_res
    veto_n = int(veto_length / veto_res) + 1

    # 1. Find large peaks in the data.
    # This will actually return big agglomerations of peaks and their tails
    peaks = strax.find_peaks(
        records,
        to_pe,
        gap_threshold=1,
        left_extension=0,
        right_extension=0,
        min_channels=100,
        min_area=area_threshold,
        result_dtype=strax.peak_dtype(n_channels=len(to_pe), n_sum_wv_samples=veto_n),
    )

    # 2a. Set 'candidate regions' at these peaks. These should:
    #  - Have a fixed maximum length (else we can't use the strax hitfinder on them)
    #  - Never extend beyond the current chunk
    #  - Do not overlap
    veto_start = peaks["time"]
    veto_end = np.clip(peaks["time"] + veto_length, None, chunk_end)
    veto_end[:-1] = np.clip(veto_end[:-1], None, veto_start[1:])

    # 2b. Convert these into strax record-like objects
    # Note the waveform is float32 though (it's a summed waveform)
    regions = np.zeros(
        len(veto_start),
        dtype=strax.interval_dtype
        + [
            ("data", (np.float32, veto_n)),
            ("baseline", np.float32),
            ("baseline_rms", np.float32),
            ("reduction_level", np.int64),
            ("record_i", np.int64),
            ("pulse_length", np.int64),
        ],
    )
    regions["time"] = veto_start
    regions["length"] = (veto_end - veto_start) // veto_n
    regions["pulse_length"] = veto_n
    regions["dt"] = veto_res

    if not len(regions):
        # No veto anywhere in this data
        return records, records[:0], np.zeros(0, strax.hit_dtype)

    # 3. Find pass_veto regios with big peaks inside the veto regions.
    # For this we compute a rough sum waveform (at low resolution,
    # without looping over the pulse data)
    rough_sum(regions, records, to_pe, veto_n, veto_res)
    if max_veto_value is not None:
        pass_veto = strax.find_hits(regions, min_amplitude=max_veto_value)
    else:
        regions["data"] /= np.max(regions["data"], axis=1)[:, np.newaxis]
        pass_veto = strax.find_hits(regions, min_amplitude=pass_veto_fraction)

    # 4. Extend these by a few samples and inverse to find veto regions
    regions["data"] = 1
    regions = strax.cut_outside_hits(
        regions, pass_veto, left_extension=pass_veto_extend, right_extension=pass_veto_extend
    )
    regions["data"] = 1 - regions["data"]
    veto = strax.find_hits(regions, min_amplitude=1)
    # Do not remove very tiny regions
    veto = veto[veto["length"] > 2 * pass_veto_extend]

    # 5. Apply the veto and return results
    veto_mask = strax.fully_contained_in(records, veto) == -1
    return tuple(list(mask_and_not(records, veto_mask)) + [veto])


@numba.njit(cache=True, nogil=True)
def rough_sum(regions, records, to_pe, n, dt):
    """Compute ultra-rough sum waveforms for regions, assuming:

     - every record is a single peak at its first sample
     - all regions have the same length and dt
    and probably not carying too much about boundaries

    """
    if not len(regions) or not len(records):
        return

    # dt and n are passed explicitly to avoid overflows/wraparounds
    # related to the small dt integer type

    peak_i = 0
    r_i = 0
    while (peak_i <= len(regions) - 1) and (r_i <= len(records) - 1):
        p = regions[peak_i]
        l = p["time"]  # noqa
        r = l + n * dt

        while True:
            if r_i > len(records) - 1:
                # Scan ahead until records contribute
                break
            t = records[r_i]["time"]
            if t >= r:
                break
            if t >= l:
                index = int((t - l) // dt)
                regions[peak_i]["data"][index] += (
                    records[r_i]["area"] * to_pe[records[r_i]["channel"]]
                )
            r_i += 1
        peak_i += 1


##
# Pulse counting
##
@export
def pulse_count_dtype(n_channels):
    # NB: don't use the dt/length interval dtype, integer types are too small
    # to contain these huge chunk-wide intervals
    return [
        (("Start time of the chunk", "time"), np.int64),
        (("End time of the chunk", "endtime"), np.int64),
        (("Number of pulses", "pulse_count"), (np.int64, n_channels)),
        (("Number of lone pulses", "lone_pulse_count"), (np.int64, n_channels)),
        (("Integral of all pulses in ADC_count x samples", "pulse_area"), (np.int64, n_channels)),
        (
            ("Integral of lone pulses in ADC_count x samples", "lone_pulse_area"),
            (np.int64, n_channels),
        ),
        (("Average baseline", "baseline_mean"), (np.int16, n_channels)),
        (("Average baseline rms", "baseline_rms_mean"), (np.float32, n_channels)),
    ]


def count_pulses(records, n_channels):
    """Return array with one element, with pulse count info from records."""
    if len(records):
        result = np.zeros(1, dtype=pulse_count_dtype(n_channels))
        _count_pulses(records, n_channels, result)
        return result
    return np.zeros(0, dtype=pulse_count_dtype(n_channels))


NO_PULSE_COUNTS = -9999  # Special value required by average_baseline in case counts = 0


@numba.njit(cache=True, nogil=True)
def _count_pulses(records, n_channels, result):
    count = np.zeros(n_channels, dtype=np.int64)
    lone_count = np.zeros(n_channels, dtype=np.int64)
    area = np.zeros(n_channels, dtype=np.int64)
    lone_area = np.zeros(n_channels, dtype=np.int64)

    last_end_seen = 0
    next_start = 0

    # Array of booleans to track whether we are currently in a lone pulse
    # in each channel
    in_lone_pulse = np.zeros(n_channels, dtype=np.bool_)
    baseline_buffer = np.zeros(n_channels, dtype=np.float64)
    baseline_rms_buffer = np.zeros(n_channels, dtype=np.float64)
    for r_i, r in enumerate(records):
        if r_i != len(records) - 1:
            next_start = records[r_i + 1]["time"]

        ch = r["channel"]
        if ch >= n_channels:
            print("Channel:", ch)
            raise RuntimeError("Out of bounds channel in get_counts!")

        area[ch] += r["area"]  # <-- Summing total area in channel

        if r["record_i"] == 0:
            count[ch] += 1
            baseline_buffer[ch] += r["baseline"]
            baseline_rms_buffer[ch] += r["baseline_rms"]

            if r["time"] > last_end_seen and r["time"] + r["pulse_length"] * r["dt"] < next_start:
                # This is a lone pulse
                lone_count[ch] += 1
                in_lone_pulse[ch] = True
                lone_area[ch] += r["area"]
            else:
                in_lone_pulse[ch] = False

            last_end_seen = max(last_end_seen, r["time"] + r["pulse_length"] * r["dt"])

        elif in_lone_pulse[ch]:
            # This is a subsequent fragment of a lone pulse
            lone_area[ch] += r["area"]

    res = result[0]
    res["pulse_count"][:] = count[:]
    res["lone_pulse_count"][:] = lone_count[:]
    res["pulse_area"][:] = area[:]
    res["lone_pulse_area"][:] = lone_area[:]
    means = baseline_buffer / count
    means[np.isnan(means)] = NO_PULSE_COUNTS
    res["baseline_mean"][:] = means[:]
    res["baseline_rms_mean"][:] = (baseline_rms_buffer / count)[:]


##
# Misc
##
@export
@numba.njit(cache=True, nogil=True)
def mask_and_not(x, mask):
    return x[mask], x[~mask]


@export
@numba.njit(cache=True, nogil=True)
def channel_split(rr, first_other_ch):
    """Return."""
    return mask_and_not(rr, rr["channel"] < first_other_ch)


@export
def check_overlaps(records, n_channels):
    """Raise a ValueError if any of the pulses in records overlap.

    Assumes records is already sorted by time.

    """
    last_end = np.zeros(n_channels, dtype=np.int64)
    channel, time = _check_overlaps(records, last_end)
    if channel != -9999:
        raise ValueError(
            f"Bad data! In channel {channel}, a pulse starts at {time}, "
            "BEFORE the previous pulse in that same channel ended "
            f"(at {last_end[channel]})"
        )


@numba.njit(cache=True, nogil=True)
def _check_overlaps(records, last_end):
    for r in records:
        if r["time"] < last_end[r["channel"]]:
            return r["channel"], r["time"]
        last_end[r["channel"]] = strax.endtime(r)
    return -9999, -9999



import numpy as np
import strax
import straxen

export, __all__ = strax.exporter()


@export
class EventAreaPerChannel(strax.Plugin):
    """Simple plugin that provides area per channel for main and alternative S1/S2 in the event."""

    depends_on = ("event_basics", "peaks")
    provides = "event_area_per_channel"
    __version__ = "0.2.0"

    compressor = "zstd"

    n_top_pmts = straxen.URLConfig(default=straxen.n_top_pmts, type=int, help="Number of top PMTs")

    def infer_dtype(self):
        # setting data type from peak dtype
        pfields_ = self.deps["peaks"].dtype_for("peaks").fields
        # populating data type
        infoline = {
            "s1": "main S1",
            "s2": "main S2",
            "alt_s1": "alternative S1",
            "alt_s2": "alternative S2",
        }
        dtype = []
        # populating APC
        ptypes = ["s1", "s2", "alt_s1", "alt_s2"]
        for type_ in ptypes:
            dtype += [
                (
                    (f"Area per channel for {infoline[type_]}", f"{type_}_area_per_channel"),
                    pfields_["area_per_channel"][0],
                )
            ]
            dtype += [
                (
                    (f"Length of the interval in samples for {infoline[type_]}", f"{type_}_length"),
                    pfields_["length"][0],
                )
            ]
            dtype += [
                (
                    (f"Width of one sample for {infoline[type_]} [ns]", f"{type_}_dt"),
                    pfields_["dt"][0],
                )
            ]
        dtype += strax.time_fields
        return dtype

    def compute(self, events, peaks):
        result = np.zeros(len(events), self.dtype)
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)

        split_peaks = strax.split_by_containment(peaks, events)
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            for type_ in ["s1", "s2", "alt_s1", "alt_s2"]:
                type_index = event[f"{type_}_index"]
                if type_index != -1:
                    result[f"{type_}_area_per_channel"][event_i] = sp["area_per_channel"][
                        type_index
                    ]
                    result[f"{type_}_length"][event_i] = sp["length"][type_index]
                    result[f"{type_}_dt"][event_i] = sp["dt"][type_index]
        return result



import typing
import matplotlib.pyplot as plt
import numpy as np
import pandas
import pymongo
import strax
import straxen
import utilix


@straxen.mini_analysis()
def daq_plot(
    context, figsize=(14, 15), lower_panel_height=6, group_by="link", vmin=None, vmax=None, **kwargs
):
    """Plot with peak, records and records sorted by "link" or "ADC ID" (other items are also
    possible as long as it is in the channel map)."""
    f, axes = plt.subplots(
        3, 1, figsize=figsize, gridspec_kw={"height_ratios": [1, 1, lower_panel_height]}
    )

    # Panel 1, the peaks
    plt.sca(axes[0])
    plt.title("Peaks")
    context.plot_peaks(**kwargs, single_figure=False)
    xlim = plt.xlim()
    plt.xticks(rotation=0)
    plt.grid("y")

    # Panel 2, the records where we keep the order of the records/channel number
    plt.sca(axes[1])
    plt.title("Records (by channel number)")
    context.plot_records_matrix(**kwargs, vmin=vmin, vmax=vmax, single_figure=False)
    plt.xticks(rotation=0)
    plt.grid("x")
    plt.xlim(*xlim)

    # Use a grouping argument to group the channels by.
    plt.sca(axes[2])
    plt.title(f"Records (by {group_by})")
    context.plot_records_matrix(
        **kwargs, vmin=vmin, vmax=vmax, group_by=group_by, single_figure=False
    )
    plt.xlim(*xlim)
    plt.grid()


def _get_daq_config(
    run_id: str,
    config_name: str = "daq_config",
    run_collection: typing.Optional[pymongo.collection.Collection] = None,
) -> dict:
    """Query the runs database for the config of the daq during this run.

    Either use the context of the runs collection.

    """
    if run_collection is None:
        if not straxen.utilix_is_configured():
            raise NotImplementedError("Only works with the runs-database")
        run_collection = utilix.rundb.xent_collection()
    daq_doc = run_collection.find_one({"number": int(run_id)}, projection={config_name: 1})
    if daq_doc is None or config_name not in daq_doc:
        raise ValueError(f"Requested {config_name} does not exist")
    return daq_doc[config_name]


def _board_to_host_link(daq_config: dict, board: int, add_crate=True) -> str:
    """Parse the daq-config to get the host, link and crate."""
    for bdoc in daq_config["boards"]:
        try:
            if int(bdoc["board"]) == board:
                res = f"{bdoc['host']}_link{bdoc['link']}"
                if add_crate:
                    res += f"_crate{bdoc['crate']}"
                return res
        except KeyError:
            raise ValueError(f"Invalid DAQ config {daq_config} or board {board}")
    # This happens if the board is not in the channel map which might
    # happen for very old runs.
    return "unknown"


def _get_cable_map(name: str = "xenonnt_cable_map.csv") -> pandas.DataFrame:
    """Download the cable map and return as a pandas dataframe."""
    down = utilix.mongo_storage.MongoDownloader()
    cable_map = down.download_single(name)
    cable_map = pandas.read_csv(cable_map)
    return cable_map


def _group_channels_by_index(
    cable_map: pandas.DataFrame,
    group_by: str = "ADC ID",
) -> typing.Tuple[np.ndarray, np.ndarray]:
    """Parse the cable map, return the labels where each of the channels is mapped to as well as an
    array that can be used to map each of the channels maps to the labels."""
    idx = np.arange(straxen.n_tpc_pmts)
    idx_seen = 0
    labels = []
    for selection in np.unique(cable_map[group_by].values):
        selected_channels = cable_map[cable_map[group_by] == selection]["PMT Location"]
        selected_channels = np.array(selected_channels)
        n_sel = len(selected_channels)

        idx[idx_seen : idx_seen + n_sel] = selected_channels
        labels += [selection] * n_sel
        idx_seen += n_sel
    return np.array(labels), idx


def group_by_daq(run_id, group_by: str):
    """From the channel map, get the mapping of channel number -> group by."""
    cable_map = _get_cable_map()
    if group_by == "link":
        labels, idx = _group_channels_by_index(cable_map, group_by="ADC ID")
        daq_config = _get_daq_config(run_id)
        labels = [_board_to_host_link(daq_config, label) for label in labels]
        labels = np.array(labels)
        order = strax.stable_argsort(labels)
        return labels[order], idx[order]
    else:
        return _group_channels_by_index(cable_map, group_by=group_by)



import strax
import straxen
from straxen.plugins.defaults import NV_HIT_DEFAULTS


export, __all__ = strax.exporter()


@export
class nVETOPulseProcessing(strax.Plugin):
    """
    nVETO equivalent of pulse processing. The following steps are
    applied:

        1. Flip, baseline and integrate waveforms.
        2. Find hits and apply ZLE
        3. Remove empty fragments.
    """

    __version__ = "0.0.8"

    parallel = "process"
    rechunk_on_save = False
    compressor = "zstd"
    save_when = strax.SaveWhen.TARGET

    depends_on = "raw_records_coin_nv"
    provides = "records_nv"
    data_kind = "records_nv"

    save_outside_hits_nv = straxen.URLConfig(
        default=NV_HIT_DEFAULTS["save_outside_hits_nv"],
        track=True,
        infer_type=False,
        help="Save (left, right) samples besides hits; cut the rest",
    )

    hit_min_amplitude_nv = straxen.URLConfig(
        infer_type=False,
        default=NV_HIT_DEFAULTS["hit_min_amplitude_nv"],
        track=True,
        help=(
            "Minimum hit amplitude in ADC counts above baseline. "
            "Specify as a tuple of length n_nveto_pmts, or a number."
        ),
    )

    baseline_samples_nv = straxen.URLConfig(
        infer_type=False,
        default=26,
        track=True,
        help="Number of samples to use at the start of the pulse to determine the baseline",
    )

    def setup(self):
        self.baseline_samples = self.baseline_samples_nv
        self.hit_thresholds = self.hit_min_amplitude_nv

    def infer_dtype(self):
        record_length = strax.record_length_from_dtype(
            self.deps["raw_records_coin_nv"].dtype_for("raw_records_coin_nv")
        )
        dtype = strax.record_dtype(record_length)
        return dtype

    def compute(self, raw_records_coin_nv):
        # Do not trust in DAQ + strax.baseline to leave the
        # out-of-bounds samples to zero.
        r = strax.raw_to_records(raw_records_coin_nv)
        del raw_records_coin_nv

        r = strax.sort_by_time(r)
        strax.zero_out_of_bounds(r)
        strax.baseline(r, baseline_samples=self.baseline_samples, flip=True)

        strax.integrate(r)

        strax.zero_out_of_bounds(r)

        hits = strax.find_hits(r, min_amplitude=self.hit_thresholds)

        le, re = self.save_outside_hits_nv
        r = strax.cut_outside_hits(r, hits, left_extension=le, right_extension=re)
        strax.zero_out_of_bounds(r)

        return r



import numpy as np
import strax
import straxen
from straxen.plugins.peaklets.peaklets import Peaklets
from straxen.plugins.defaults import HE_PREAMBLE

export, __all__ = strax.exporter()


@export
class PeakletsHighEnergy(Peaklets):
    __doc__ = HE_PREAMBLE + (Peaklets.__doc__ or "")
    depends_on = "records_he"
    provides = "peaklets_he"
    data_kind = "peaklets_he"
    __version__ = "0.1.0"
    child_plugin = True
    save_when = strax.SaveWhen.TARGET

    n_he_pmts = straxen.URLConfig(
        track=False, default=752, infer_type=False, help="Maximum channel of the he channels"
    )

    he_channel_offset = straxen.URLConfig(
        track=False, default=500, infer_type=False, help="Minimum channel number of the he channels"
    )

    le_to_he_amplification = straxen.URLConfig(
        default=20,
        track=True,
        infer_type=False,
        help="Difference in amplification between low energy and high energy channels",
    )

    peak_min_pmts_he = straxen.URLConfig(
        default=2,
        infer_type=False,
        child_option=True,
        parent_option_name="peak_min_pmts",
        track=True,
        help="Minimum number of contributing PMTs needed to define a peak",
    )

    saturation_correction_on_he = straxen.URLConfig(
        default=False,
        infer_type=False,
        child_option=True,
        parent_option_name="saturation_correction_on",
        track=True,
        help="On off switch for saturation correction for High Energy channels",
    )

    hit_min_amplitude_he = straxen.URLConfig(
        default=(
            "list-to-array://"
            "xedocs://hit_thresholds"
            "?as_list=True"
            "&sort=pmt"
            "&attr=value"
            "&detector=tpc_he"
            "&run_id=plugin.run_id"
            "&version=ONLINE"
        ),
        help="Minimum hit amplitude in ADC counts above baseline. "
        "Specify as a tuple of length n_tpc_pmts, or a number,"
        'or a string like "pmt_commissioning_initial" which means calling'
        "hitfinder_thresholds.py",
        track=True,
        infer_type=False,
    )

    # We cannot, we only have the top array, so should not.
    store_data_top = False

    @property
    def n_tpc_pmts(self):
        # Have to hack the url config to avoid nasty numba errors for the main Peaklets plugin
        return self.n_he_pmts

    def infer_dtype(self):
        return strax.peak_dtype(n_channels=self.n_he_pmts, store_data_top=self.store_data_top)

    def setup(self):
        self.to_pe = self.gain_model
        buffer_pmts = np.zeros(self.he_channel_offset)
        self.to_pe = np.concatenate((buffer_pmts, self.to_pe))
        self.to_pe *= self.le_to_he_amplification
        self.hit_thresholds = self.hit_min_amplitude_he
        self.channel_range = self.channel_map["he"]
        self._tight_coincidence_window_left = self.tight_coincidence_window_left
        self._tight_coincidence_window_right = self.tight_coincidence_window_right

    def compute(self, records_he, start, end):
        result = super().compute(records_he, start, end)
        return result["peaklets"]



import numpy as np
import strax


class EventSEScore(strax.Plugin):
    """This plugin is designed to calculate the single electron rate density for main and alt S2 in
    events.

    References:
        * v0.0.0: xenon:xenonnt:analysis:hot_spot_cut_summary
        * v0.1.0: xenon:xenonnt:ac:sr1:hotspot_veto_cut:wimp_roi

    """

    __version__ = "0.1.0"
    depends_on = ("event_basics", "peak_se_score")
    provides = "event_se_score"

    def infer_dtype(self):
        dtype = []
        for s_i in [1, 2]:
            for main_or_alt, description in zip(["", "alt_"], ["main", "alternate"]):
                dtype += [
                    (
                        (
                            f"Neiboring single-electron score of {description} S{s_i} ",
                            f"{main_or_alt}s{s_i}_se_score",
                        ),
                        np.float32,
                    ),
                ]
        dtype += strax.time_fields
        return dtype

    def compute(self, events, peaks):
        split_peaks = strax.split_by_containment(peaks, events)
        result = np.zeros(len(events), self.dtype)

        # 1. Assign peaks features to main S2 and alt S2 in the event
        for event_i, (event, sp) in enumerate(zip(events, split_peaks)):
            for type_ in ["s1", "alt_s1", "s2", "alt_s2"]:
                type_index = event[f"{type_}_index"]
                if type_index != -1:
                    result[f"{type_}_se_score"][event_i] = sp["se_score"][type_index]

        # 2. Set time and endtime for events
        result["time"] = events["time"]
        result["endtime"] = strax.endtime(events)
        return result



import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import strax
import straxen
from mpl_toolkits.axes_grid1 import inset_locator

from .daq_waveforms import group_by_daq
from .records_matrix import DEFAULT_MAX_SAMPLES

export, __all__ = strax.exporter()


@straxen.mini_analysis()
def plot_waveform(
    context,
    deep=False,
    show_largest=100,
    figsize=None,
    max_samples=DEFAULT_MAX_SAMPLES,
    ignore_max_sample_warning=True,
    cbar_loc="lower right",
    lower_panel_height=2,
    **kwargs,
):
    """Plot the sum waveform and optionally per-PMT waveforms.

    :param deep: If True, show per-PMT waveform matrix under sum waveform. If 'raw', use raw_records
        instead of records to do so.
    :param show_largest: Show only the largest show_largest peaks.
    :param figsize: Matplotlib figure size for the plot.
    :param cbar_loc: location of the intensity color bar. Set to None to omit it altogether.
    :param lower_panel_height: Height of the lower panel in terms of the height of the upper panel.

    """
    if figsize is None:
        figsize = (10, 6 if deep else 4)

    if not deep:
        context.plot_peaks(**kwargs, show_largest=show_largest, figsize=figsize)

    else:
        f, axes = plt.subplots(
            2,
            1,
            constrained_layout=True,
            figsize=figsize,
            gridspec_kw={"height_ratios": [1, lower_panel_height]},
        )

        plt.sca(axes[0])
        context.plot_peaks(**kwargs, show_largest=show_largest, single_figure=False, xaxis=False)

        plt.sca(axes[1])
        context.plot_records_matrix(
            **kwargs,
            cbar_loc=cbar_loc,
            max_samples=max_samples,
            ignore_max_sample_warning=ignore_max_sample_warning,
            raw=deep == "raw",
            single_figure=False,
        )

        plt.subplots_adjust(hspace=0)


@straxen.mini_analysis(
    requires=("peaks", "peak_basics"), default_time_selection="touching", warn_beyond_sec=60
)
def plot_peaks(
    peaks,
    seconds_range,
    t_reference,
    show_largest=100,
    single_figure=True,
    figsize=(10, 4),
    xaxis=True,
):
    if single_figure:
        plt.figure(figsize=figsize)
    plt.axhline(0, c="k", alpha=0.2)

    peaks = peaks[strax.stable_argsort(-peaks["area"])[:show_largest]]
    peaks = strax.sort_by_time(peaks)

    color_map = {1: "blue", 2: "green"}
    for p in peaks:
        plot_peak(p, t0=t_reference, color=color_map.get(p["type"], "gray"))

    if xaxis == "since_start":
        seconds_range_xaxis(seconds_range, t0=seconds_range[0])
    elif xaxis:
        seconds_range_xaxis(seconds_range)
        plt.xlim(*seconds_range)
    else:
        plt.xticks([])
        plt.xlim(*seconds_range)
    plt.ylabel("Intensity [PE/ns]")
    if single_figure:
        plt.tight_layout()


@straxen.mini_analysis(
    requires=("peaks", "peak_basics"), default_time_selection="touching", warn_beyond_sec=60
)
def plot_hit_pattern(
    peaks,
    seconds_range,
    t_reference,
    axes=None,
    vmin=None,
    log_scale=False,
    label=None,
    single_figure=False,
    figsize=(10, 4),
):
    if single_figure:
        plt.figure(figsize=figsize)
    if len(peaks) > 1:
        print(f"warning showing total area of {len(peaks)} peaks")
    straxen.plot_pmts(
        np.sum(peaks["area_per_channel"], axis=0),
        axes=axes,
        vmin=vmin,
        log_scale=log_scale,
        label=label,
    )


@straxen.mini_analysis()
def plot_records_matrix(
    context,
    run_id,
    seconds_range,
    cbar_loc="upper right",
    raw=False,
    single_figure=True,
    figsize=(10, 4),
    group_by=None,
    max_samples=DEFAULT_MAX_SAMPLES,
    ignore_max_sample_warning=False,
    vmin=None,
    vmax=None,
    **kwargs,
):
    if seconds_range is None:
        raise ValueError(
            "You must pass a time selection (e.g. seconds_range) to plot_records_matrix."
        )

    if single_figure:
        plt.figure(figsize=figsize, constrained_layout=True)

    f = context.raw_records_matrix if raw else context.records_matrix

    wvm, ts, ys = f(
        run_id,
        max_samples=max_samples,
        ignore_max_sample_warning=ignore_max_sample_warning,
        **kwargs,
    )
    if group_by is not None:
        ylabs, wvm_mask = group_by_daq(run_id, group_by)
        wvm = wvm[:, wvm_mask]
        plt.ylabel(group_by)
    else:
        plt.ylabel("Channel number")

    # extract min and max from kwargs or set defaults
    if vmin is None:
        vmin = min(0.1 * wvm.max(), 1e-2)
    if vmax is None:
        vmax = wvm.max()

    plt.pcolormesh(
        ts,
        ys,
        wvm.T,
        norm=matplotlib.colors.LogNorm(
            vmin=vmin,
            vmax=vmax,
        ),
        cmap=plt.cm.inferno,
    )
    plt.xlim(*seconds_range)

    ax = plt.gca()
    if group_by is not None:
        # Do some magic to convert all the labels to an integer that
        # allows for remapping of the y labels to whatever is provided
        # in the "ylabs", otherwise matplotlib shows nchannels different
        # labels in the case of strings.
        # Make a dict that converts the label to an int
        int_labels = {h: i for i, h in enumerate(set(ylabs))}
        mask = np.ones(len(ylabs), dtype=bool)
        # If the label (int) is different wrt. its neighbour, show it
        mask[1:] = np.abs(np.diff([int_labels[y] for y in ylabs])) > 0
        # Only label the selection
        ax.set_yticks(np.arange(len(ylabs))[mask])
        ax.set_yticklabels(ylabs[mask])
    plt.xlabel("Time [s]")

    if cbar_loc is not None:
        # Create a white box to place the color bar in
        # See https://stackoverflow.com/questions/18211967
        bbox = inset_locator.inset_axes(ax, width="20%", height="22%", loc=cbar_loc)
        _ = [bbox.spines[k].set_visible(False) for k in bbox.spines]
        bbox.patch.set_facecolor((1, 1, 1, 0.9))
        bbox.set_xticks([])
        bbox.set_yticks([])

        # Create the actual color bar
        cax = inset_locator.inset_axes(bbox, width="90%", height="20%", loc="upper center")
        plt.colorbar(cax=cax, label="Intensity [PE/ns]", orientation="horizontal")
        cax.xaxis.set_major_formatter(matplotlib.ticker.FormatStrFormatter("%g"))

    plt.sca(ax)


def seconds_range_xaxis(seconds_range, t0=None):
    """Make a pretty time axis given seconds_range."""
    plt.xlim(*seconds_range)
    ax = plt.gca()
    ax.ticklabel_format(useOffset=False)
    xticks = plt.xticks()[0]
    if not len(xticks):
        return

    # Format the labels
    # I am not very proud of this code...
    def chop(x):
        return np.floor(x).astype(np.int64)

    if t0 is None:
        xticks_ns = np.round(xticks * int(1e9)).astype(np.int64)
    else:
        xticks_ns = np.round((xticks - xticks[0]) * int(1e9)).astype(np.int64)
    sec = chop(xticks_ns // int(1e9))
    ms = chop((xticks_ns % int(1e9)) // int(1e6))
    us = chop((xticks_ns % int(1e6)) // int(1e3))
    samples = chop((xticks_ns % int(1e3)) // 10)

    labels = [str(sec[i]) for i in range(len(xticks))]
    print_ns = np.any(samples != samples[0])
    print_us = print_ns | np.any(us != us[0])
    print_ms = print_us | np.any(ms != ms[0])
    if print_ms and t0 is None:
        labels = [l + f".{ms[i]:03}" for i, l in enumerate(labels)]
        if print_us:
            labels = [l + r" $\bf{" + f"{us[i]:03}" + "}$" for i, l in enumerate(labels)]
            if print_ns:
                labels = [l + f" {samples[i]:02}0" for i, l in enumerate(labels)]
        plt.xticks(ticks=xticks, labels=labels, rotation=90)
    else:
        labels = list(chop((xticks_ns // 10) * 10))
        labels[-1] = ""
        plt.xticks(ticks=xticks, labels=labels, rotation=0)
    if t0 is None:
        plt.xlabel("Time since run start [sec]")
    else:
        plt.xlabel("Time [ns]")


def plot_peak(p, t0=None, center_time=True, **kwargs):
    x, y = time_and_samples(p, t0=t0)
    kwargs.setdefault("linewidth", 1)

    # Plot waveform
    plt.plot(x, y, drawstyle="steps-pre", **kwargs)
    if "linewidth" in kwargs:
        del kwargs["linewidth"]
    kwargs["alpha"] = kwargs.get("alpha", 1) * 0.2
    plt.fill_between(x, 0, y, step="pre", linewidth=0, **kwargs)

    # Mark extent with thin black line
    plt.plot([x[0], x[-1]], [y.max(), y.max()], c="k", alpha=0.3, linewidth=1)

    # Mark center time with thin black line
    if center_time:
        if t0 is None:
            t0 = p["time"]
        ct = (p["center_time"] - t0) / int(1e9)
        plt.axvline(ct, c="k", alpha=0.4, linewidth=1, linestyle="--")


def time_and_samples(p, t0=None):
    """Return (x, y) numpy arrays for plotting the waveform data in p using 'steps-pre'.

    Where x is the time since t0 in seconds (or another time_scale), and y is intensity in PE / ns.
    :param p: Peak or other similar strax data type
    :param t0: Zero of time in ns since unix epoch

    """
    n = p["length"]
    if t0 is None:
        t0 = p["time"]
    x = ((p["time"] - t0) + np.arange(n + 1) * p["dt"]) / int(1e9)
    y = p["data"][:n] / p["dt"]
    return x, np.concatenate([[y[0]], y])



# temporarily ignored, need to see whther the redefine is necessary
# mypy: disable-error-code="no-redef"

from . import defaults
from .defaults import *

# TPC chain first, order matters!
from . import raw_records
from .raw_records import *

from . import records
from .records import *

from . import peaklets
from .peaklets import *

from . import merged_s2s
from .merged_s2s import *

from . import peaks
from .peaks import *

from . import events
from .events import *

from .aqmon_hits import *
from . import aqmon_hits

from . import veto_intervals
from .veto_intervals import *

from . import online_peak_monitor
from .online_peak_monitor import *

from . import individual_peak_monitor
from .individual_peak_monitor import *

# NV chain
from . import raw_records_coin_nv
from .raw_records_coin_nv import *

from . import records_nv
from .records_nv import *

from . import hitlets_nv
from .hitlets_nv import *

from . import events_nv
from .events_nv import *

from . import online_monitor_nv
from .online_monitor_nv import *

# MV chain
from . import records_mv
from .records_mv import *

from . import hitlets_mv
from .hitlets_mv import *

from . import events_mv
from .events_mv import *

from . import online_monitor_mv
from .online_monitor_mv import *

# HE
from . import records_he
from .records_he import *

from . import merged_s2s_he
from .merged_s2s_he import *

from . import peaklets_he
from .peaklets_he import *

from .peaks_he import *
from . import peaks_he

# Misc
from . import afterpulses
from .afterpulses import *

from . import led_cal
from .led_cal import *

from . import ref_mon_nv
from .ref_mon_nv import *

from . import gps_syncing
from .gps_syncing import *



import strax
from straxen.plugins.peaks._peak_positions_base import PeakPositionsBase

export, __all__ = strax.exporter()


@export
class MergedS2Positions(PeakPositionsBase):

    __version__ = "0.0.0"
    child_plugin = True
    provides = "merged_s2_positions"
    depends_on = (
        "merged_s2_positions_mlp",
        "merged_s2_positions_cnf",
    )

    def compute(self, merged_s2s):
        return super().compute(merged_s2s)
